###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import re
import os
import iso8601
import humanize
import uuid
import threading
import pprint
from datetime import datetime
from collections import OrderedDict
from pathlib import Path

import cbor2
import nacl
import numpy as np

from autobahn import wamp, util
from autobahn.twisted.wamp import ApplicationSession
from autobahn.util import utcnow
from autobahn.wamp.types import RegisterOptions, CallDetails, PublishOptions
from autobahn.wamp.request import Registration

from twisted.internet.defer import inlineCallbacks
from twisted.internet.task import LoopingCall
from twisted.internet.threads import deferToThread

import txaio
import treq
import zlmdb
from txaio import time_ns

from crossbar._util import hlid, hl, hlval
from crossbar.node.main import _get_versions
from crossbar.common import checkconfig
from crossbar.common.key import _read_node_key, _read_release_key, _write_node_key, _parse_node_key

from cfxdb.globalschema import GlobalSchema
from cfxdb.mrealmschema import MrealmSchema
from crossbar.master.mrealm.mrealm import MrealmManager, ManagementRealm, Node
from crossbar.master.node.user import UserManager
from cfxdb.user import User, UserRole, UserMrealmRole
from cfxdb.usage import MasterNodeUsage
from cfxdb.log import MWorkerLog

_CFC_DOMAIN = 'crossbarfabriccenter.domain.'
_CFC_MREALM = 'crossbarfabriccenter.mrealm.'
_CFC_USER = 'crossbarfabriccenter.user.'


class DomainManager(object):
    """
    Global domain backend.
    """
    def __init__(self, session, db, schema):
        self.log = session.log
        self._db = db
        self._schema = schema

        self._session = session

    def register(self, session, prefix, options):
        return session.register(self, prefix=prefix, options=options)

    @wamp.register(None)
    def get_status(self, details=None):
        """
        Get global (domain) realm status.

        :procedure: ``crossbarfabriccenter.domain.get_status``

        :param details: Call details.
        :type details: :class:`autobahn.wamp.types.CallOptions`

        :returns: Global system status information.
        :rtype: dict
        """
        assert isinstance(details, CallDetails)

        self.log.debug('{klass}.get_status(details={details})', klass=self.__class__.__name__, details=details)

        now = utcnow()
        uptime_secs = (iso8601.parse_date(now) - iso8601.parse_date(self._session._started)).total_seconds()
        uptime_secs_str = humanize.naturaldelta(uptime_secs)
        res = {
            'type': 'domain',
            'realm': self._session._realm,
            'now': utcnow(),
            'started': self._session._started,
            'uptime': uptime_secs_str,
            'tick': self._session._tick
        }
        return res

    @wamp.register(None)
    def get_version(self, details=None):
        """
        Returns CFC software stack version information.

        :procedure: ``crossbarfabriccenter.domain.get_version``

        :param details: Call details.
        :type details: :class:`autobahn.wamp.types.CallOptions`

        :return: Information on software stack versions.
        :rtype: dict
        """
        assert isinstance(details, CallDetails)

        self.log.debug('{klass}.get_version(details={details})', klass=self.__class__.__name__, details=details)

        # FIXME
        from twisted.internet import reactor
        versions = _get_versions(reactor)

        return versions.marshal()

    @wamp.register(None)
    def get_license(self, details=None):
        """
        Returns CFC software stack license information.

        :procedure: ``crossbarfabriccenter.domain.get_license``

        :param details: Call details.
        :type details: :class:`autobahn.wamp.types.CallOptions`

        :return: License information, including enabled features and limits.
        :rtype: dict
        """
        assert isinstance(details, CallDetails)

        self.log.debug('{klass}.get_license(details={details})', klass=self.__class__.__name__, details=details)

        # FIXME: check blockchain for license information
        license = {
            'product': 'crossbar-free-tier',
            'description': 'crossbar free usage tier including 5 managed nodes.',
            'terms': 'https://crossbario.com/license',
            'features': {
                'xbr': False
            },
            'limits': {
                'concurrent-nodes': 5
            }
        }
        return license


class DomainController(ApplicationSession):
    """
    Main CFC backend for a domain (instance of CFC, all CFC nodes connected to one controller node / database).

    There is exactly one instance running per master node of a domain.
    """
    log = txaio.make_logger()

    def _initialize(self):
        # self.config.controller
        # self.config.shared

        # comes from "crossbar/master/node/config.json"
        assert self.config.extra
        cbdir = self.config.extra['cbdir']

        # Release (public) key
        self._release_pubkey_hex = _read_release_key()['hex']

        # Node key
        self._node_key_hex = _read_node_key(cbdir, private=True)['hex']
        self._node_key = nacl.signing.SigningKey(self._node_key_hex, encoder=nacl.encoding.HexEncoder)

        # Metering knobs - FIXME: read and honor all knobs
        self._meterurl = self.config.extra.get('metering', {}).get('submit', {}).get('url', '${CROSSBAR_METERING_URL}')

        self._meterurl = checkconfig.maybe_from_env('metering.submit.url', self._meterurl)

        self.log.debug(
            'Initialized {klass} from ComponentConfig.extra [node_key="{node_key}", meter_url="{meter_url}"]',
            klass=self.__class__.__name__,
            meter_url=self._meterurl,
            node_key=self._node_key_hex)

        # auto-create default management realm, auto-pair nodes to default mrealm from watching directories
        # for new node public keys. configuration example:
        #
        #       "auto_default_mrealm": {
        #           "enabled": true,
        #           "watch_to_pair": [],
        #           "write_pairing_file": true
        #        }
        #
        self._auto_default_mrealm = self.config.extra.get('auto_default_mrealm', False)

        # create database and attach tables to database slots
        #
        config = self.config.extra.get('database', {})

        dbpath = config.get('path', '.db-controller')
        assert type(dbpath) == str
        dbpath = os.path.join(cbdir, dbpath)

        maxsize = config.get('maxsize', 128 * 2**20)
        assert type(maxsize) == int
        # allow maxsize 128kiB to 128GiB
        assert maxsize >= 128 * 1024 and maxsize <= 128 * 2**30

        # create database and attach tables to database slots
        #
        self.db = zlmdb.Database(dbpath=dbpath, maxsize=maxsize, readonly=False, sync=True)
        self.db.__enter__()
        self.schema = GlobalSchema.attach(self.db)

        self.log.debug('{klass} global database opened [dbpath={dbpath}, maxsize={maxsize}]',
                       klass=self.__class__.__name__,
                       dbpath=hlid(dbpath),
                       maxsize=hlid(maxsize))

    async def onJoin(self, details):

        self._initialize()

        # check if magic special builtin "superuser" exists
        #
        with self.db.begin(write=True) as txn:
            self._superuser_authid = 'superuser'
            self._superuser_oid = self.schema.idx_users_by_email[txn, self._superuser_authid]
            if not self._superuser_oid:
                self._superuser_oid = uuid.uuid4()
                user = User()
                user.oid = self._superuser_oid
                user.email = self._superuser_authid
                user.pubkey = None
                user.registered = datetime.utcnow()
                self.schema.users[txn, user.oid] = user
                self.log.info(
                    hl('SUPERUSER created and stored in database (oid={}, email={})'.format(user.oid, user.email),
                       color='green',
                       bold=True))
            else:
                self.log.info(
                    hl('SUPERUSER already exists in database (oid={})'.format(self._superuser_oid),
                       color='green',
                       bold=True))

        # check if default management realm exists
        #
        self._watch_and_pair_lc = None
        self._watch_and_pair_count = None
        self._default_mrealm_oid = None
        self._default_mrealm_enabled = self._auto_default_mrealm.get('enabled', False)
        if self._default_mrealm_enabled:
            self.log.info('{action}', action=hl('Default management realm enabled', color='green', bold=True))
        else:
            self.log.info('{action}', action=hl('Default management realm disabled', color='red', bold=True))
        if self._default_mrealm_enabled:
            # the default management realm (if exists) has always the name "default"!
            mrealm_name = 'default'

            with self.db.begin(write=True) as txn:
                self._default_mrealm_oid = self.schema.idx_mrealms_by_name[txn, mrealm_name]
                if self._default_mrealm_oid:
                    self.log.debug('ok, default management realm already exists ({oid})',
                                   oid=hlid(self._default_mrealm_oid))

                    # This should not happen - but better verify that the owner of the default management realm
                    # is the current superuser.
                    # FIXME: merely logging a warning is likely not enough to address this!
                    #
                    mrealm = self.schema.mrealms[txn, self._default_mrealm_oid]
                    if mrealm.owner != self._superuser_oid:
                        self.log.warn(
                            '\n\nWARNING: default management realm not owned by SUPERUSER! [owner={owner}, superuser={superuser}]\n\n',
                            superuser=hlid(self._superuser_oid),
                            owner=hlid(mrealm.owner))
                else:
                    # the OID of a newly created default management realm varies locally
                    self._default_mrealm_oid = uuid.uuid4()

                    new_mrealm = ManagementRealm()
                    new_mrealm.oid = self._default_mrealm_oid
                    new_mrealm.label = 'default mrealm'
                    new_mrealm.description = 'Default management realm (automatically pre-created)'
                    new_mrealm.tags = []
                    new_mrealm.name = mrealm_name
                    new_mrealm.created = datetime.utcnow()
                    new_mrealm.owner = self._superuser_oid
                    new_mrealm.cf_router = 'cfrouter1'
                    new_mrealm.cf_container = 'cfcontainer1'

                    # store new management realm
                    self.schema.mrealms[txn, new_mrealm.oid] = new_mrealm

                    # store roles for user that created the management realm
                    roles = UserMrealmRole([UserRole.OWNER, UserRole.ADMIN, UserRole.USER, UserRole.GUEST])
                    self.schema.users_mrealm_roles[txn, (self._superuser_oid, new_mrealm.oid)] = roles

                    self.log.info('{action} [oid={oid}]',
                                  action=hl('Default management realm created', color='red', bold=True),
                                  oid=hlid(new_mrealm.oid))

            self._watch_to_pair = self._auto_default_mrealm.get('watch_to_pair', None)
            if self._watch_to_pair:
                self.log.info('{action}', action=hl('Watch-to-pair enabled', color='green', bold=True))
            else:
                self.log.info('{action}', action=hl('Watch-to-pair disabled', color='red', bold=True))
            if self._watch_to_pair:
                self._watch_to_pair = checkconfig.maybe_from_env('auto_default_mrealm.watch_to_pair',
                                                                 self._watch_to_pair,
                                                                 hide_value=False)
                if self._watch_to_pair:
                    self._watch_to_pair = Path(self._watch_to_pair)
                if self._watch_to_pair and self._watch_to_pair.is_dir():
                    node_dir_pat = self._auto_default_mrealm.get('watch_to_pair_pattern', None)
                    if node_dir_pat:
                        node_dir_pat = re.compile(node_dir_pat)

                    # FIXME: read from config
                    follow_symlinks = True

                    @inlineCallbacks
                    def watch_and_pair():
                        self.log.debug('{klass}::watch_and_pair[counter={cnt}, watch_to_pair="{watch_to_pair}"]',
                                       cnt=hlval(self._watch_and_pair_count),
                                       watch_to_pair=self._watch_to_pair,
                                       klass=self.__class__.__name__)
                        try:
                            self._watch_and_pair_count += 1

                            if node_dir_pat:
                                node_dirs = [
                                    x.name for x in os.scandir(self._watch_to_pair)
                                    if x.is_dir(follow_symlinks=follow_symlinks) and node_dir_pat.match(x.name)
                                ]
                            else:
                                node_dirs = [
                                    x.name for x in os.scandir(self._watch_to_pair)
                                    if x.is_dir(follow_symlinks=follow_symlinks)
                                ]

                            pubkeys = []
                            for node_dir in node_dirs:
                                # r = root, d = directories, f = files
                                for r, d, f in os.walk(os.path.join(self._watch_to_pair, node_dir)):
                                    for file in f:
                                        if file == 'key.pub':
                                            node_key_file = os.path.join(r, file)
                                            if os.path.isfile(node_key_file):
                                                node_key_tags = _parse_node_key(node_key_file)
                                                node_key_hex = node_key_tags['public-key-ed25519']
                                                node_id = node_key_tags.get('node-authid', None)
                                                cluster_ip = node_key_tags.get('node-cluster-ip', None)
                                                pubkeys.append((r, node_key_hex, node_id, cluster_ip))

                            self.log.debug(
                                '{klass}::watch_and_pair: found {cnt} directories (matching), with {cntk} node keys scanned ..',
                                cnt=hl(len(node_dirs)),
                                cntk=hl(len(pubkeys)),
                                klass=self.__class__.__name__)

                            if pubkeys:
                                # determine actually new pubkeys
                                new_pubkeys = []
                                with self.db.begin() as txn:
                                    for cbdir, pubkey, node_id, cluster_ip in pubkeys:
                                        node_oid = self.schema.idx_nodes_by_pubkey[txn, pubkey]
                                        if not node_oid:
                                            new_pubkeys.append((cbdir, pubkey, node_id, cluster_ip))
                                        else:
                                            node = self.schema.nodes[txn, node_oid]
                                            self.log.debug(
                                                '{klass}::watch_and_pair: node with pubkey {pubkey} already paired as node {node} to mrealm {mrealm}',
                                                klass=self.__class__.__name__,
                                                pubkey=hlid(pubkey),
                                                node=hlid(node.oid),
                                                mrealm=hlid(node.mrealm_oid))

                                self.log.debug(
                                    '{klass}::watch_and_pair: scanned ({count}) directory "{watch_to_pair}" for node auto-pairing, found {cnt_new_keys} new (in {cnt_keys} total) keys',
                                    count=hlval(self._watch_and_pair_count),
                                    watch_to_pair=hlid(self._watch_to_pair),
                                    cnt_keys=hlval(len(pubkeys)),
                                    cnt_new_keys=hlval(len(new_pubkeys)),
                                    klass=self.__class__.__name__)

                                # store all new pubkeys
                                for cbdir, pubkey, node_id, cluster_ip in new_pubkeys:
                                    node = Node()
                                    node.oid = uuid.uuid4()
                                    node.pubkey = pubkey
                                    node.cluster_ip = cluster_ip

                                    # auto-pair newly discovered node to the default management realm, and owned by superuser
                                    node.owner_oid = self._superuser_oid
                                    node.mrealm_oid = self._default_mrealm_oid
                                    if node_id:
                                        node.authid = node_id
                                    else:
                                        node.authid = 'node-{}'.format(str(node.oid)[:8])
                                    node.authextra = {
                                        'node_oid': str(node.oid),
                                        'cluster_ip': cluster_ip,
                                        'mrealm_oid': str(node.mrealm_oid),
                                    }

                                    # store paired node in database
                                    with self.db.begin(write=True) as txn:
                                        self.schema.nodes[txn, node.oid] = node

                                    if self._auto_default_mrealm.get('write_pairing_file', False):
                                        management_url = self._auto_default_mrealm.get('management_url', None)
                                        if management_url:
                                            management_url = checkconfig.maybe_from_env(
                                                'auto_default_mrealm.management_url', management_url, hide_value=False)

                                        activation_code = None
                                        if self._auto_default_mrealm.get('include_activation_code', False):
                                            activation_code = util.generate_activation_code()
                                            # FIXME: store activation & check later

                                        activation_file = os.path.join(cbdir, 'key.activate')
                                        if not os.path.exists(activation_file):
                                            file_tags = OrderedDict([
                                                ('created-at', utcnow()),
                                                ('management-url', management_url or 'wss://master.xbr.network/ws'),
                                                ('management-realm', 'default'),
                                                ('management-realm-oid', str(node.mrealm_oid)),
                                                ('node-oid', str(node.oid)),
                                                ('node-cluster-ip', node.cluster_ip),
                                                ('node-authid', str(node.authid)),
                                                ('activation-code', activation_code),
                                                ('public-key-ed25519', pubkey),
                                            ])
                                            file_msg = 'Crossbar.io node activation\n\n'
                                            try:
                                                _write_node_key(activation_file, file_tags, file_msg)
                                            except OSError as e:
                                                self.log.warn(
                                                    '{klass}::watch_and_pair: failed to write {action} to {activation_file} ({err})',
                                                    klass=self.__class__.__name__,
                                                    action=hl('Node activation file', color='red', bold=True),
                                                    activation_file=hlval(activation_file),
                                                    err=str(e))
                                            else:
                                                self.log.info(
                                                    '{klass}::watch_and_pair: {action} written to {activation_file}',
                                                    klass=self.__class__.__name__,
                                                    action=hl('Node activation file', color='red', bold=True),
                                                    activation_file=hlval(activation_file))
                                        else:
                                            self.log.warn(
                                                '{klass}::watch_and_pair: skipped writing {action} to {activation_file} (path already exists)',
                                                klass=self.__class__.__name__,
                                                action=hl('Node activation file', color='red', bold=True),
                                                activation_file=hlval(activation_file))

                                    topic = 'crossbarfabriccenter.mrealm.on_node_paired'
                                    payload = node.marshal()
                                    yield self.publish(topic, payload, options=PublishOptions(acknowledge=True))

                                    self.log.info(
                                        '{klass}::watch_and_pair: {action} with pubkey={pubkey}, oid={node_oid}, authid={authid} to default management realm {mrealm}!',
                                        klass=self.__class__.__name__,
                                        action=hl('Auto-paired node', color='red', bold=True),
                                        node_oid=hlid(node.oid),
                                        authid=hlid(node.authid),
                                        pubkey=hlval(node.pubkey),
                                        mrealm=hlid(node.mrealm_oid))
                            else:
                                self.log.debug(
                                    '{klass}::watch_and_pair: scanned ({count}) directory "{watch_to_pair}" for node auto-pairing: no nodes found',
                                    klass=self.__class__.__name__,
                                    count=hlval(self._watch_and_pair_count),
                                    watch_to_pair=hlid(self._watch_to_pair))
                        except:
                            self.log.failure()

                    self._watch_and_pair_lc = LoopingCall(watch_and_pair)
                    self._watch_and_pair_count = 0
                    self._watch_and_pair_lc.start(10)
                else:
                    if self._watch_to_pair:
                        self.log.warn('skipping to watch "{watch_to_pair}" for node auto-pairing - not a directory!',
                                      watch_to_pair=self._watch_to_pair.absolute())
                    else:
                        self.log.warn('skipping to watch for node auto-pairing - no directory configured!')

        # initialize management backends
        #
        self._domain_mgr = DomainManager(self, self.db, self.schema)
        self._user_mgr = UserManager(self, self.db, self.schema)
        self._mrealm_mgr = MrealmManager(self, self.db, self.schema)

        domains = (
            (_CFC_DOMAIN, self._domain_mgr.register),
            (_CFC_USER, self._user_mgr.register),
            (_CFC_MREALM, self._mrealm_mgr.register),
        )  # yapf: disable
        for topic, procedure in domains:
            results = await procedure(self, prefix=topic, options=RegisterOptions(details_arg='details'))
            for reg in results:
                if type(reg) == Registration:
                    self.log.debug('Registered CFC API <{proc}>', proc=reg.procedure)
                else:
                    self.log.error('Error: <{}>'.format(reg.value.args[0]))

        # start master heartbeat loop ..
        #
        self._tick = 1

        @inlineCallbacks
        def tick():
            # process regular master activities
            started = time_ns()

            # watch master database fill grade
            dbstats = self.db.stats()
            free = dbstats['free']
            used = dbstats['current_size']
            if free < 0.01:
                self.log.error('Global master database full! Initiating EMERGENCY SHUTDOWN .. [{free}% free]',
                               free=round(free * 100., 2))

                # if we don't have at least 1% free space, immediately shutdown the whole node so we don't
                # run into "db full" later possibly half way through the body of a management procedure
                yield self.config.controller.call('crossbar.shutdown')
            elif free < 0.1:
                self.log.warn('Global master database almost full: only {free}% free space left!',
                              free=round(free * 100., 2))

            # >>> BEGIN of master heartbeat loop tasks

            # 1) aggregate and store usage metering records
            cnt_new = None
            if True:
                try:
                    cnt_new = yield self._do_metering(started)
                except:
                    self.log.failure()

            # 2) submit usage meterings records to metering service
            if self._meterurl:
                if cnt_new:
                    try:
                        yield self._submit_metering(started)
                    except:
                        self.log.failure()
            else:
                self.log.warn('Skipping to submit metering records - no metering URL set!')

            # >>> END of master heartbeat loop tasks

            # publish master heartbeat
            ticked = {'now': utcnow(), 'tick': self._tick}
            yield self.publish('{}on_tick'.format(_CFC_DOMAIN), ticked, options=PublishOptions(acknowledge=True))

            # master heartbeat loop finished!
            duration = int(round((time_ns() - started) / 1000000.))
            if duration > 500:
                self.log.warn('Master heartbeat loop iteration {tick} finished: excessive run-time of {duration} ms!',
                              tick=self._tick,
                              duration=duration)
            self.log.debug(
                'Master heartbeat loop iteration {tick} finished in {duration} ms (database {used} used, {free}% free)',
                tick=self._tick,
                free=round(free * 100., 2),
                used=humanize.naturalsize(used),
                duration=duration)

            # set next master heartbeat loop iteration sequence number
            self._tick += 1

        # FIXME: make this configurable from master node config
        master_tick_period = 300
        c = LoopingCall(tick)
        c.start(master_tick_period)
        self.log.debug('Master heartbeat loop started .. [period={period} secs]', period=master_tick_period)

        # note status started
        #
        self._started = utcnow()
        self.log.debug('Domain controller ready (realm="{realm}")!', realm=hlid(self._realm))

    def _first_metering(self, mrealm_id):
        """
        Determine timestamp of first node heartbeat from any node of a given management realm.

        :param mrealm_id:
        :return:
        """
        dbpath = os.path.join(self.config.extra['cbdir'], '.db-mrealm-{}'.format(mrealm_id))
        db = zlmdb.Database(dbpath=dbpath, readonly=False)
        schema = MrealmSchema.attach(db)
        with self.db.begin() as txn:
            with db.begin() as txn2:
                for (ts, node_id) in schema.mnode_logs.select(txn2, reverse=False, return_values=False):
                    node = self.schema.nodes[txn, node_id]
                    if node and node.mrealm_oid == mrealm_id:
                        return ts

    def _agg_metering_mnode_logs(self, from_ts, until_ts, mrealm_id, by_node=False):
        """
        Aggregate raw managed node and worker logs and store usage metering records.

        Note: This is run on a background thread!

        :param from_ts:
        :param until_ts:
        :param mrealm_id:
        :param by_node:
        :return:
        """

        dbpath = os.path.join(self.config.extra['cbdir'], '.db-mrealm-{}'.format(mrealm_id))
        db = zlmdb.Database(dbpath=dbpath, readonly=False)
        schema = MrealmSchema.attach(db)

        if by_node:
            # compute aggregate sum grouped by node_id
            res = {}
            with db.begin() as txn:
                for (ts, node_id) in schema.mnode_logs.select(txn,
                                                              from_key=(from_ts, uuid.UUID(bytes=b'\x00' * 16)),
                                                              to_key=(until_ts, uuid.UUID(bytes=b'\xff' * 16)),
                                                              return_values=False,
                                                              reverse=False):

                    rec = schema.mnode_logs[txn, (ts, node_id)]

                    if node_id not in res:
                        res[node_id] = {
                            'count': 0,
                            'nodes': 0,
                            'routers': 0,
                            'containers': 0,
                            'guests': 0,
                            'proxies': 0,
                            'marketmakers': 0,
                            'hostmonitors': 0,
                            'controllers': 0,
                        }

                    res[node_id]['count'] += 1
                    res[node_id]['nodes'] += rec.period
                    res[node_id]['routers'] += rec.routers * rec.period
                    res[node_id]['containers'] += rec.containers * rec.period
                    res[node_id]['guests'] += rec.guests * rec.period
                    res[node_id]['proxies'] += rec.proxies * rec.period
                    res[node_id]['marketmakers'] += rec.marketmakers * rec.period
                    res[node_id]['hostmonitors'] += rec.hostmonitors * rec.period
                    res[node_id]['controllers'] += rec.controllers * rec.period

        else:
            # compute aggregate sum
            res = {
                'count': 0,
                'nodes': 0,
                'routers': 0,
                'containers': 0,
                'guests': 0,
                'proxies': 0,
                'marketmakers': 0,
                'hostmonitors': 0,
                'controllers': 0,
            }
            nodes = set()
            with db.begin() as txn:
                for (ts, node_id) in schema.mnode_logs.select(txn,
                                                              from_key=(from_ts, uuid.UUID(bytes=b'\x00' * 16)),
                                                              to_key=(until_ts, uuid.UUID(bytes=b'\xff' * 16)),
                                                              return_values=False,
                                                              reverse=False):

                    rec = schema.mnode_logs[txn, (ts, node_id)]

                    if node_id not in nodes:
                        nodes.add(node_id)

                    res['count'] += 1
                    res['nodes'] += rec.period
                    res['routers'] += rec.routers * rec.period
                    res['containers'] += rec.containers * rec.period
                    res['guests'] += rec.guests * rec.period
                    res['proxies'] += rec.proxies * rec.period
                    res['marketmakers'] += rec.marketmakers * rec.period
                    res['hostmonitors'] += rec.hostmonitors * rec.period
                    res['controllers'] += rec.controllers * rec.period

        self.log.debug(
            '  Metering: aggregated node logs metering records on thread {thread_id} [mrealm_id={mrealm_id}, from_ts={from_ts}, until_ts={until_ts}]:\n{res}',
            mrealm_id=mrealm_id,
            from_ts=from_ts,
            until_ts=until_ts,
            thread_id=threading.get_ident(),
            res=pprint.pformat(res))

        return res

    def _agg_metering_mworker_logs(self, from_ts, until_ts, mrealm_id):
        """
        Aggregate raw managed node and worker logs and store usage metering records.

        Note: This is run on a background thread!

        :param from_ts:
        :param until_ts:
        :param mrealm_id:
        :param by_node:
        :return:
        """

        dbpath = os.path.join(self.config.extra['cbdir'], '.db-mrealm-{}'.format(mrealm_id))
        db = zlmdb.Database(dbpath=dbpath, readonly=False)
        schema = MrealmSchema.attach(db)

        # compute aggregate sum
        res = {
            'count': 0,
            'total': 0,
            'controllers': 0,
            'hostmonitors': 0,
            'routers': 0,
            'containers': 0,
            'guests': 0,
            'proxies': 0,
            'marketmakers': 0,
            'sessions': 0,
            'msgs_call': 0,
            'msgs_yield': 0,
            'msgs_invocation': 0,
            'msgs_result': 0,
            'msgs_error': 0,
            'msgs_publish': 0,
            'msgs_published': 0,
            'msgs_event': 0,
            'msgs_register': 0,
            'msgs_registered': 0,
            'msgs_subscribe': 0,
            'msgs_subscribed': 0,
        }
        wres = {}
        nodes = set()
        with db.begin() as txn:
            # go over all worker heartbeat records in given time interval ..
            for (ts, node_id,
                 worker_id) in schema.mworker_logs.select(txn,
                                                          from_key=(from_ts, uuid.UUID(bytes=b'\x00' * 16), ''),
                                                          to_key=(until_ts, uuid.UUID(bytes=b'\xff' * 16), ''),
                                                          return_values=False,
                                                          reverse=False):

                rec = schema.mworker_logs[txn, (ts, node_id, worker_id)]

                assert rec.period

                # set of nodes we encountered in heartbeats in the time interval
                if node_id not in nodes:
                    nodes.add(node_id)

                # type name (str) of the worker
                worker_type = MWorkerLog.WORKER_TYPENAMES[rec.type]

                # increment used "worker seconds"
                # FIXME: cleanup this hack ..
                if worker_type.endswith('y'):
                    # proxy -> proxies
                    res['{}ies'.format(worker_type[:-1])] += rec.period
                else:
                    res['{}s'.format(worker_type)] += rec.period

                # increment processed records
                res['count'] += 1

                # for workers of type "router", we compute additional statistics:
                if worker_type == 'router':
                    wkey = (node_id, worker_id)
                    if wkey not in wres:
                        wres[wkey] = {
                            # session seconds
                            'sessions': 0,

                            # minimum number of WAMP messages per type
                            'msgs_call_min': 0,
                            'msgs_yield_min': 0,
                            'msgs_invocation_min': 0,
                            'msgs_result_min': 0,
                            'msgs_error_min': 0,
                            'msgs_publish_min': 0,
                            'msgs_published_min': 0,
                            'msgs_event_min': 0,
                            'msgs_register_min': 0,
                            'msgs_registered_min': 0,
                            'msgs_subscribe_min': 0,
                            'msgs_subscribed_min': 0,

                            # maximum number of WAMP messages per type
                            'msgs_call_max': 0,
                            'msgs_yield_max': 0,
                            'msgs_invocation_max': 0,
                            'msgs_result_max': 0,
                            'msgs_error_max': 0,
                            'msgs_publish_max': 0,
                            'msgs_published_max': 0,
                            'msgs_event_max': 0,
                            'msgs_register_max': 0,
                            'msgs_registered_max': 0,
                            'msgs_subscribe_max': 0,
                            'msgs_subscribed_max': 0,
                        }

                    # increment used "session seconds" (seconds of connected clients)
                    wres[wkey]['sessions'] += rec.router_sessions * rec.period

                    if rec.recv_call > wres[wkey]['msgs_call_max']:
                        wres[wkey]['msgs_call_max'] = rec.recv_call
                    if not wres[wkey]['msgs_call_min'] or rec.recv_call < wres[wkey]['msgs_call_min']:
                        wres[wkey]['msgs_call_min'] = rec.recv_call

                    if rec.recv_yield > wres[wkey]['msgs_yield_max']:
                        wres[wkey]['msgs_yield_max'] = rec.recv_yield
                    if not wres[wkey]['msgs_yield_min'] or rec.recv_yield < wres[wkey]['msgs_yield_min']:
                        wres[wkey]['msgs_yield_min'] = rec.recv_yield

                    if rec.sent_invocation > wres[wkey]['msgs_invocation_max']:
                        wres[wkey]['msgs_invocation_max'] = rec.sent_invocation
                    if not wres[wkey]['msgs_invocation_min'] or rec.sent_invocation < wres[wkey]['msgs_invocation_min']:
                        wres[wkey]['msgs_invocation_min'] = rec.sent_invocation

                    # FIXME
                    # if rec.sent_error > res[wkey]['msgs_error_max']:
                    #     res[wkey]['msgs_error_max'] = rec.sent_error
                    # if not res[wkey]['msgs_error_min'] or rec.sent_error < res[wkey]['msgs_error_min']:
                    #     res[wkey]['msgs_error_min'] = rec.sent_error

                    if rec.sent_result > wres[wkey]['msgs_result_max']:
                        wres[wkey]['msgs_result_max'] = rec.sent_result
                    if not wres[wkey]['msgs_result_min'] or rec.sent_result < wres[wkey]['msgs_result_min']:
                        wres[wkey]['msgs_result_min'] = rec.sent_result

                    if rec.recv_publish > wres[wkey]['msgs_publish_max']:
                        wres[wkey]['msgs_publish_max'] = rec.recv_publish
                    if not wres[wkey]['msgs_publish_min'] or rec.recv_publish < wres[wkey]['msgs_publish_min']:
                        wres[wkey]['msgs_publish_min'] = rec.recv_publish

                    if rec.sent_published > wres[wkey]['msgs_published_max']:
                        wres[wkey]['msgs_published_max'] = rec.sent_published
                    if not wres[wkey]['msgs_published_min'] or rec.sent_published < wres[wkey]['msgs_published_min']:
                        wres[wkey]['msgs_published_min'] = rec.sent_published

                    if rec.sent_event > wres[wkey]['msgs_event_max']:
                        wres[wkey]['msgs_event_max'] = rec.sent_event
                    if not wres[wkey]['msgs_event_min'] or rec.sent_event < wres[wkey]['msgs_event_min']:
                        wres[wkey]['msgs_event_min'] = rec.sent_event

                    if rec.recv_register > wres[wkey]['msgs_register_max']:
                        wres[wkey]['msgs_register_max'] = rec.recv_register
                    if not wres[wkey]['msgs_register_min'] or rec.recv_register < wres[wkey]['msgs_register_min']:
                        wres[wkey]['msgs_register_min'] = rec.recv_register

                    if rec.sent_registered > wres[wkey]['msgs_registered_max']:
                        wres[wkey]['msgs_registered_max'] = rec.sent_registered
                    if not wres[wkey]['msgs_registered_min'] or rec.sent_registered < wres[wkey]['msgs_registered_min']:
                        wres[wkey]['msgs_registered_min'] = rec.sent_registered

                    if rec.recv_subscribe > wres[wkey]['msgs_subscribe_max']:
                        wres[wkey]['msgs_subscribe_max'] = rec.recv_subscribe
                    if not wres[wkey]['msgs_subscribe_min'] or rec.recv_subscribe < wres[wkey]['msgs_subscribe_min']:
                        wres[wkey]['msgs_subscribe_min'] = rec.recv_subscribe

                    if rec.sent_subscribed > wres[wkey]['msgs_subscribed_max']:
                        wres[wkey]['msgs_subscribed_max'] = rec.sent_subscribed
                    if not wres[wkey]['msgs_subscribed_min'] or rec.sent_subscribed < wres[wkey]['msgs_subscribed_min']:
                        wres[wkey]['msgs_subscribed_min'] = rec.sent_subscribed

        res['nodes'] = len(nodes)

        for wkey in wres:
            res['sessions'] += wres[wkey]['sessions']
            res['msgs_call'] += wres[wkey]['msgs_call_max'] - wres[wkey]['msgs_call_min']
            res['msgs_yield'] += wres[wkey]['msgs_yield_max'] - wres[wkey]['msgs_yield_min']
            res['msgs_invocation'] += wres[wkey]['msgs_invocation_max'] - wres[wkey]['msgs_invocation_min']
            res['msgs_result'] += wres[wkey]['msgs_result_max'] - wres[wkey]['msgs_result_min']
            res['msgs_error'] += wres[wkey]['msgs_error_max'] - wres[wkey]['msgs_error_min']
            res['msgs_publish'] += wres[wkey]['msgs_publish_max'] - wres[wkey]['msgs_publish_min']
            res['msgs_published'] += wres[wkey]['msgs_published_max'] - wres[wkey]['msgs_published_min']
            res['msgs_event'] += wres[wkey]['msgs_event_max'] - wres[wkey]['msgs_event_min']
            res['msgs_register'] += wres[wkey]['msgs_register_max'] - wres[wkey]['msgs_register_min']
            res['msgs_registered'] += wres[wkey]['msgs_registered_max'] - wres[wkey]['msgs_registered_min']
            res['msgs_subscribe'] += wres[wkey]['msgs_subscribe_max'] - wres[wkey]['msgs_subscribe_min']
            res['msgs_subscribed'] += wres[wkey]['msgs_subscribed_max'] - wres[wkey]['msgs_subscribed_min']

        self.log.debug(
            '  Metering: aggregated {cnt_records} records from mworker_logs: {cnt_nodes} nodes, {cnt} metering records, thread {thread_id} [mrealm_id={mrealm_id}, from_ts={from_ts}, until_ts={until_ts}]',
            cnt=len(res),
            cnt_nodes=res['nodes'],
            cnt_records=res['count'],
            mrealm_id=mrealm_id,
            from_ts=from_ts,
            until_ts=until_ts,
            thread_id=threading.get_ident())

        return res

    @inlineCallbacks
    def _do_metering(self, started):
        self.log.debug('Usage metering: aggregating heartbeat records .. [started="{started}", thread_id={thread_id}]',
                       started=np.datetime64(started, 'ns'),
                       thread_id=threading.get_ident())

        # FIXME: make this tunable from the master node config
        agg_mins = 5

        # determine intervals of (from_ts, until_ts) for which to compute usage data from aggregate raw heartbeat data
        intervals = []
        mrealm_ids = []
        with self.db.begin() as txn:
            mrealm_ids.extend(self.schema.mrealms.select(txn, return_values=False))

            last_ts = None
            for (ts, _) in self.schema.usage.select(txn, reverse=True, limit=1, return_values=False):
                last_ts = ts

            if not last_ts:
                for mrealm_id in mrealm_ids:
                    ts = self._first_metering(mrealm_id)
                    if not last_ts or ts < last_ts:
                        last_ts = ts
                if not last_ts:
                    last_ts = np.datetime64(np.datetime64(time_ns(), 'ns') - np.timedelta64(agg_mins, 'm'), 'ns')

                last_ts = np.datetime64(last_ts.astype('datetime64[m]'), 'ns')
                self.log.debug('Usage metering: first metering timestamp set to "{last_ts}"', last_ts=last_ts)
            else:
                self.log.debug('Usage metering: last metering timestamp stored is "{last_ts}"', last_ts=last_ts)

            until_ts = np.datetime64(last_ts + np.timedelta64(agg_mins, 'm'), 'ns')

            while until_ts < np.datetime64(time_ns(), 'ns'):
                intervals.append((last_ts, until_ts))
                self.log.debug('Usage metering: aggregation interval ("{last_ts}", "{until_ts}") appended',
                               last_ts=last_ts,
                               until_ts=until_ts)

                last_ts = until_ts
                until_ts = np.datetime64(last_ts + np.timedelta64(agg_mins, 'm'), 'ns')

        cnt_new = 0

        # iterate over intervals to be aggregated ..
        if intervals:
            self.log.debug(
                'Usage metering: {cnt_intervals} intervals collected - now aggregating on background threads ..',
                cnt_intervals=len(intervals))

            # iterate over intervals ..
            for from_ts, until_ts in intervals:

                # .. and all management realms ..
                # FIXME: aggregate all mrealms for a given interval in one go
                for mrealm_id in mrealm_ids:
                    try:
                        # aggregate all _worker_ heartbeat data for interval (from_ts, until_ts) and mrealm_id
                        mrealm_res = yield deferToThread(self._agg_metering_mworker_logs, from_ts, until_ts, mrealm_id)

                        self.log.debug('Usage metering: worker aggregate result\n{mrealm_node_res}',
                                       mrealm_node_res=pprint.pformat(mrealm_res))

                        # aggregate all _node_ heartbeat data for interval (from_ts, until_ts) and mrealm_id
                        mrealm_node_res = yield deferToThread(self._agg_metering_mnode_logs, from_ts, until_ts,
                                                              mrealm_id)

                        self.log.debug('Usage metering: node aggregate result\n{mrealm_node_res}',
                                       mrealm_node_res=pprint.pformat(mrealm_node_res))
                    except:
                        self.log.failure()
                    else:
                        # overwrite "nodes" with what the node-based aggregate says (which is correct):
                        mrealm_res['nodes'] = mrealm_node_res['nodes']

                        for key in ['routers', 'guests', 'containers', 'proxies', 'marketmakers', 'controllers']:
                            if mrealm_res[key] != mrealm_node_res[key]:
                                self.log.warn(
                                    'Usage metering: node/worker aggregate for worker type "{key}" differ on worker seconds: {worker_res} != {node_res} (worker-based / node-based aggregate)',
                                    key=key,
                                    worker_res=mrealm_res[key],
                                    node_res=mrealm_node_res[key])

                        mrealm_res['timestamp'] = int(until_ts)
                        mrealm_res['timestamp_from'] = int(from_ts)
                        mrealm_res['mrealm_id'] = str(mrealm_id)
                        mrealm_res['seq'] = self._tick

                        # set master (!) node public key
                        mrealm_res['pubkey'] = self._node_key.verify_key.encode(encoder=nacl.encoding.RawEncoder)

                        # usage records start in status "RECEIVED"
                        mrealm_res['status'] = 1
                        mrealm_res['processed'] = time_ns()

                        # not used here
                        # mrealm_res['sent'] = None
                        # mrealm_res['status_message'] = None

                        # parse the raw dict into object
                        usage = MasterNodeUsage.parse(mrealm_res)

                        # finally, store new usage record in database
                        with self.db.begin(write=True) as txn:
                            self.schema.usage[txn, (until_ts, mrealm_id)] = usage

                        # publish usage metering event
                        topic = '{}on_meter_usage'.format(_CFC_DOMAIN)
                        yield self.publish(topic,
                                           str(mrealm_id),
                                           str(from_ts),
                                           str(until_ts),
                                           usage.marshal(),
                                           options=PublishOptions(acknowledge=True))

                        self.log.debug(
                            'Usage metering: aggregated and stored period from {from_ts} to {until_ts} ({duration}) usage metering data for mrealm "{mrealm_id}"',
                            mrealm_id=mrealm_id,
                            duration=str(np.timedelta64(until_ts - from_ts, 's')),
                            from_ts=from_ts,
                            until_ts=until_ts)
                        self.log.debug('Usage metering data:\n{usage}', usage=pprint.pformat(usage.marshal()))

                        cnt_new += 1
        else:
            self.log.debug('Usage metering: no new intervals to aggregate.')

        self.log.debug('Usage metering: finished aggregating [{cnt_new} intervals stored]', cnt_new=cnt_new)

        return cnt_new

    @inlineCallbacks
    def _submit_metering(self, started, filter_status=[1], limit=None):
        self.log.debug('Usage metering: submitting metering records .. [started="{started}"]',
                       started=np.datetime64(started, 'ns'))

        # collects keys in table "schema.usage" for metering records to be submitted
        keys = []

        # we only submit meterings up to 24h old
        from_ts = np.datetime64(np.datetime64(time_ns(), 'ns') - np.timedelta64(24, 'h'), 'ns')
        from_key = (from_ts, uuid.UUID(bytes=b'\x00' * 16))

        with self.db.begin() as txn:
            for rec in self.schema.usage.select(txn, reverse=False, return_keys=False, from_key=from_key, limit=limit):
                if rec.status in filter_status:
                    keys.append((rec.timestamp, rec.mrealm_id))

        tried = 0
        success = 0
        failed = 0

        # submit each metering record (here, sequentially)
        for key in keys:
            # fetch the record we want to submit
            with self.db.begin() as txn:
                rec = self.schema.usage[txn, key]

            self.log.debug('Usage metering: submitting metering record to "{url}"\n{rec}',
                           rec=pprint.pformat(rec.marshal()),
                           url=self._meterurl)

            # the master node public key
            verify_key = self._node_key.verify_key.encode(encoder=nacl.encoding.RawEncoder)

            # serialize metering data
            raw_data = cbor2.dumps(rec.marshal())

            # sign metering data with master node (private) key
            signed_msg = self._node_key.sign(raw_data)

            # POST message body is concatenation of verify key and signed message:
            data = verify_key + signed_msg

            self.log.debug('HTTP/POST: verify_key={lvk} data={ld} raw_data={lrd} signed_msg={lsm}',
                           lvk=len(verify_key),
                           ld=len(data),
                           lrd=len(raw_data),
                           lsm=len(signed_msg))

            tried += 1
            metering_id = None

            try:
                # issue the actual (outgoing) HTTP/POST request (with a 5s timeout) ..
                response = yield treq.post(self._meterurl, data=data, timeout=5)

                # .. and receive response body
                rdata = yield treq.content(response)

                if response.code != 200:
                    raise Exception('metering denied by metering service: HTTP response {}, "{}")'.format(
                        response.code, rdata))

            except Exception as e:
                # eg "twisted.internet.error.ConnectionRefusedError"
                # self.log.failure()
                rec.status = 3
                rec.status_message = 'failed to submit metering record: {}'.format(e)
                rec.processed = np.datetime64(time_ns(), 'ns')
                failed += 1
                self.log.warn('Usage metering: failed to submit metering record for "{timestamp}" - "{errmsg}"',
                              timestamp=rec.timestamp,
                              errmsg=str(e))
            else:
                try:
                    metering_id = uuid.UUID(bytes=rdata)
                except Exception as e:
                    rec.status = 3
                    rec.status_message = 'invalid response from metering service: {}'.format(e)
                    rec.processed = np.datetime64(time_ns(), 'ns')
                    failed += 1
                    self.log.log_failure()
                else:
                    rec.status = 2
                    rec.processed = np.datetime64(time_ns(), 'ns')
                    rec.metering_id = metering_id
                    success += 1
                    self.log.debug(
                        'Usage metering: metering record for "{timestamp}" successfully submitted [meterurl="{meterurl}", metering_id="{metering_id}"]',
                        meterurl=self._meterurl,
                        timestamp=rec.timestamp,
                        metering_id=metering_id,
                        response_length=len(rdata))

            with self.db.begin(write=True) as txn:
                self.schema.usage[txn, key] = rec

            self.log.debug(
                'Usage metering: metering record for "{timestamp}" processed with new status {status} [metering_id="{metering_id}"].',
                timestamp=rec.timestamp,
                metering_id=metering_id,
                status=rec.status)

        self.log.debug(
            'Usage metering: finished submitting metering records [tried={tried}, success={success}, failed={failed}]',
            tried=tried,
            success=success,
            failed=failed)



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

from autobahn import xbr
from crossbar._version import __version__, __build__

import txaio

txaio.use_twisted()

from crossbar.master.personality import Personality  # noqa

__all__ = ('__version__', '__build__', 'Personality', 'xbr')
__doc__ = Personality.DESC



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################


def native_string(string):
    """
    Make C{string} be the type of C{str}, decoding with ASCII if required.
    """
    if isinstance(string, bytes):
        return string.decode('ascii')
    else:
        raise ValueError("This is already a native string.")



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################


class RouterOptions(object):
    """
    Router options for creating routers.
    """

    URI_CHECK_LOOSE = "loose"
    URI_CHECK_STRICT = "strict"

    def __init__(self, uri_check=None, event_dispatching_chunk_size=None):
        """

        :param uri_check: Method which should be applied to check WAMP URIs.
        :type uri_check: str
        """
        self.uri_check = uri_check or RouterOptions.URI_CHECK_STRICT
        self.event_dispatching_chunk_size = event_dispatching_chunk_size or 100

    def __str__(self):
        return ("RouterOptions(uri_check = {0}, "
                "event_dispatching_chunk_size = {1})".format(
                    self.uri_check,
                    self.event_dispatching_chunk_size,
                ))


class NotAttached(RuntimeError):
    """
    Internal error: session not attached to router.
    """



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from typing import Dict, Type

from crossbar.router.auth.pending import PendingAuth  # noqa
from crossbar.router.auth.anonymous import PendingAuthAnonymous  # noqa
from crossbar.router.auth.anonymous import PendingAuthAnonymousProxy  # noqa
from crossbar.router.auth.wampcra import PendingAuthWampCra  # noqa
from crossbar.router.auth.ticket import PendingAuthTicket  # noqa
from crossbar.router.auth.tls import PendingAuthTLS  # noqa
from crossbar.router.auth.scram import PendingAuthScram  # noqa

AUTHMETHODS = set([
    'ticket',
    'wampcra',
    'tls',
    'cryptosign',
    'cryptosign-proxy',
    'cookie',
    'anonymous',
    'anonymous-proxy',
    'scram',
])

# map of authmethod name to processor class
# note that not all of AUTHMETHODS need to have an
# entry here .. eg when dependencies are missing
AUTHMETHOD_MAP: Dict[str, Type[PendingAuth]] = {
    'anonymous': PendingAuthAnonymous,
    'anonymous-proxy': PendingAuthAnonymousProxy,
    'ticket': PendingAuthTicket,
    'wampcra': PendingAuthWampCra,
    'tls': PendingAuthTLS,
    'cookie': PendingAuth,
    'scram': PendingAuthScram,
}

AUTHMETHOD_PROXY_MAP: Dict[str, Type[PendingAuth]] = {
    'anonymous-proxy': PendingAuthAnonymousProxy,
}

try:
    import nacl  # noqa
    HAS_CRYPTOSIGN = True
except ImportError:
    HAS_CRYPTOSIGN = False

__all__ = [
    'AUTHMETHODS',
    'AUTHMETHOD_MAP',
    'HAS_CRYPTOSIGN',
    'PendingAuth',
    'PendingAuthWampCra',
    'PendingAuthScram',
    'PendingAuthTicket',
    'PendingAuthTLS',
]

if HAS_CRYPTOSIGN:
    from crossbar.router.auth.cryptosign import PendingAuthCryptosign  # noqa
    from crossbar.router.auth.cryptosign import PendingAuthCryptosignProxy  # noqa
    __all__.append('PendingAuthCryptosign')
    __all__.append('PendingAuthCryptosignProxy')
    AUTHMETHOD_MAP['cryptosign'] = PendingAuthCryptosign
    AUTHMETHOD_MAP['cryptosign-proxy'] = PendingAuthCryptosignProxy
    AUTHMETHOD_PROXY_MAP['cryptosign-proxy'] = PendingAuthCryptosignProxy



###############################################################################
#
# Crossbar.io Shell
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import os
import shlex
import sys
import six
from collections import defaultdict

import click
import click._bashcomplete
import click.parser

from prompt_toolkit import prompt
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.styles import style_from_pygments_dict
from pygments.token import Token

from autobahn.wamp.exception import ApplicationError

from crossbar.shell.util import style_error


def _get_bottom_toolbar_tokens(cli):
    return [(Token.Toolbar, ' This is a toolbar. ')]


_style = style_from_pygments_dict({
    Token.Toolbar: '#ffffff bg:#333333',
})


class InternalCommandException(Exception):
    pass


class ExitReplException(InternalCommandException):
    pass


_internal_commands = dict()


def _register_internal_command(names, target, description=None):
    if not hasattr(target, '__call__'):
        raise ValueError('Internal command must be a callable')

    if isinstance(names, six.string_types):
        names = [names]
    elif not isinstance(names, (list, tuple)):
        raise ValueError('"names" must be a string or a list / tuple')

    for name in names:
        _internal_commands[name] = (target, description)


def _get_registered_target(name, default=None):
    target_info = _internal_commands.get(name)
    if target_info:
        return target_info[0]
    return default


def _exit_internal():
    raise ExitReplException()


def _help_internal():
    formatter = click.HelpFormatter()
    formatter.write_heading('REPL help')
    formatter.indent()
    with formatter.section('External Commands'):
        formatter.write_text('prefix external commands with "!"')
    with formatter.section('Internal Commands'):
        formatter.write_text('prefix internal commands with ":"')
        info_table = defaultdict(list)  # type: ignore
        for mnemonic, target_info in six.iteritems(_internal_commands):
            info_table[target_info[1]].append(mnemonic)
        formatter.write_dl((', '.join((':{0}'.format(mnemonic) for mnemonic in sorted(mnemonics))), description)
                           for description, mnemonics in six.iteritems(info_table))
    return formatter.getvalue()


_register_internal_command(['q', 'quit', 'exit'], _exit_internal, 'exits the repl')
_register_internal_command(['?', 'h', 'help'], _help_internal, 'displays general help information')


class ClickCompleter(Completer):
    def __init__(self, cli):
        self.cli = cli

    def get_completions(self, document, complete_event=None):
        # Code analogous to click._bashcomplete.do_complete

        try:
            args = shlex.split(document.text_before_cursor)
        except ValueError:
            # Invalid command, perhaps caused by missing closing quotation.
            return

        cursor_within_command = \
            document.text_before_cursor.rstrip() == document.text_before_cursor

        if args and cursor_within_command:
            # We've entered some text and no space, give completions for the
            # current word.
            incomplete = args.pop()
        else:
            # We've not entered anything, either at all or for the current
            # command, so give all relevant completions for this context.
            incomplete = ''

        # FIXME
        _bc = click._bashcomplete  # type: ignore
        ctx = _bc.resolve_ctx(self.cli, '', args)
        if ctx is None:
            return

        cmds = []
        c = ctx
        while c:
            cmds.append(c.command.name)
            c = c.parent
        cmds.reverse()

        # print(cmds)
        # if ctx.parent:
        #    print('COMMAND: ', ctx.parent.command.name)
        # pprint(dir(ctx.command))
        # print(document.get_word_before_cursor())
        # print(document.get_word_before_cursor(WORD=True))

        choices = []
        for param in ctx.command.params:
            if not isinstance(param, click.Option):
                continue
            for options in (param.opts, param.secondary_opts):
                for o in options:
                    choices.append(Completion(o, -len(incomplete), display_meta=param.help))

        if isinstance(ctx.command, click.MultiCommand):
            for name in ctx.command.list_commands(ctx):
                command = ctx.command.get_command(ctx, name)  # type: ignore
                choices.append(Completion(name, -len(incomplete), display_meta=getattr(command, 'short_help')))

        for item in choices:
            if item.text.startswith(incomplete):
                yield item


def continuation_tokens(cli, width):
    " The continuation: display dots before all the following lines. "

    # (make sure that the width of the continuation does not exceed the given
    # width. -- It is the prompt that decides the width of the left margin.)
    return [(Token, '.' * (width - 1) + ' ')]


async def repl(old_ctx,
               prompt_kwargs=None,
               allow_system_commands=True,
               allow_internal_commands=True,
               once=False,
               get_bottom_toolbar_tokens=_get_bottom_toolbar_tokens,
               get_prompt_tokens=None,
               style=_style):
    """
    Start an interactive shell. All subcommands are available in it.

    :param old_ctx: The current Click context.
    :param prompt_kwargs: Parameters passed to
        :py:func:`prompt_toolkit.shortcuts.prompt`.

    If stdin is not a TTY, no prompt will be printed, but only commands read
    from stdin.

    """
    # parent should be available, but we're not going to bother if not
    group_ctx = old_ctx.parent or old_ctx
    group = group_ctx.command
    isatty = sys.stdin.isatty()

    # Delete the REPL command from those available, as we don't want to allow
    # nesting REPLs (note: pass `None` to `pop` as we don't want to error if
    # REPL command already not present for some reason).
    repl_command_name = old_ctx.command.name
    available_commands = group_ctx.command.commands
    available_commands.pop(repl_command_name, None)

    if isatty:
        prompt_kwargs = prompt_kwargs or {}
        if not get_prompt_tokens:
            prompt_kwargs.setdefault('message', u'>> ')
        history = prompt_kwargs.pop('history', None) \
            or InMemoryHistory()
        completer = prompt_kwargs.pop('completer', None) \
            or ClickCompleter(group)

        def get_command():
            return prompt(
                completer=completer,
                history=history,
                # patch_stdout=False,
                # https://github.com/jonathanslenders/python-prompt-toolkit/blob/master/examples/get-multiline-input.py
                # multiline=True,
                # get_continuation_tokens=continuation_tokens,
                # get_bottom_toolbar_tokens=get_bottom_toolbar_tokens,
                # get_prompt_tokens=get_prompt_tokens,
                style=style,
                async_=True,
                **prompt_kwargs)
    else:
        get_command = sys.stdin.readline

    stopped = False
    while not stopped:
        try:
            command = await get_command()
        except KeyboardInterrupt:
            continue
        except EOFError:
            break
        finally:
            if once:
                stopped = True

        if not command:
            if isatty:
                continue
            else:
                break

        if allow_system_commands and dispatch_repl_commands(command):
            continue

        if allow_internal_commands:
            try:
                result = handle_internal_commands(command)
                if isinstance(result, six.string_types):
                    click.echo(result)
                    continue
            except ExitReplException:
                break

        args = shlex.split(command)

        try:
            with group.make_context(None, args, parent=group_ctx) as ctx:
                f = group.invoke(ctx)
                if f:
                    await f
                ctx.exit()

        except ApplicationError as e:
            click.echo(style_error(u'[{}] {}'.format(e.error, e.args[0])))

        except click.ClickException as e:
            e.show()

        except SystemExit:
            pass


def register_repl(group, name='repl'):
    """Register :func:`repl()` as sub-command *name* of *group*."""
    group.command(name=name)(click.pass_context(repl))


def dispatch_repl_commands(command):
    """Execute system commands entered in the repl.

    System commands are all commands starting with "!".

    """
    if command.startswith('!'):
        os.system(command[1:])  # nosec
        return True

    return False


def handle_internal_commands(command):
    """Run repl-internal commands.

    Repl-internal commands are all commands starting with ":".

    """
    if command.startswith(':'):
        target = _get_registered_target(command[1:], default=None)
        if target:
            return target()



###############################################################################
#
# Crossbar.io Shell
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

import os
import sys
import json
import yaml
from pprint import pformat

import click

from pygments import highlight, lexers, formatters
from pygments.token import Token
from pygments.styles import get_all_styles

from prompt_toolkit.history import FileHistory
from prompt_toolkit.styles import style_from_pygments_dict

from twisted.python.failure import Failure
from twisted.internet.task import react

import txaio
from txaio import make_logger

from autobahn.websocket.util import parse_url
from autobahn.wamp.types import ComponentConfig
from autobahn.wamp.exception import ApplicationError
from autobahn.twisted.wamp import ApplicationRunner

from crossbar.common.twisted.endpoint import _create_tls_client_context

from crossbar.shell.util import (style_crossbar, style_finished_line, style_error, style_ok, localnow)
from crossbar.shell import (client, config, userkey, __version__)

if 'CROSSBAR_FABRIC_URL' in os.environ:
    _DEFAULT_CFC_URL = os.environ['CROSSBAR_FABRIC_URL']
else:
    _DEFAULT_CFC_URL = 'wss://master.xbr.network/ws'


class WebSocketURL(click.ParamType):
    """
    WebSocket URL validator.
    """

    name = 'WebSocket URL'

    def __init__(self):
        click.ParamType.__init__(self)

    def convert(self, value, param, ctx):
        try:
            parse_url(value)
        except Exception as e:
            self.fail(style_error(str(e)))
        else:
            return value


def _prompt_for_url(yes_to_all):
    """
    Prompt user for CFC URL to create a new ~/.crossbar/config.ini file
    """
    if yes_to_all:
        value = _DEFAULT_CFC_URL
    else:
        value = click.prompt('Management service URL', type=WebSocketURL(), default=_DEFAULT_CFC_URL)
    return value


# default configuration stored in $HOME/.crossbar/config.ini
_DEFAULT_CONFIG = """[default]

url={url}
privkey=default.priv
pubkey=default.pub
"""


class Application(object):

    OUTPUT_FORMAT_PLAIN = 'plain'
    OUTPUT_FORMAT_JSON = 'json'
    OUTPUT_FORMAT_JSON_COLORED = 'json-color'
    OUTPUT_FORMAT_YAML = 'yaml'
    OUTPUT_FORMAT_YAML_COLORED = 'yaml-color'

    OUTPUT_FORMAT = [
        OUTPUT_FORMAT_PLAIN, OUTPUT_FORMAT_JSON, OUTPUT_FORMAT_JSON_COLORED, OUTPUT_FORMAT_YAML,
        OUTPUT_FORMAT_YAML_COLORED
    ]

    OUTPUT_VERBOSITY_SILENT = 'silent'
    OUTPUT_VERBOSITY_NORMAL = 'normal'
    OUTPUT_VERBOSITY_VERBOSE = 'verbose'
    OUTPUT_VERBOSITY_EXTENDED = 'extended'
    OUTPUT_VERBOSITY_RESULT_ONLY = 'results-only'

    OUTPUT_VERBOSITY = [
        OUTPUT_VERBOSITY_SILENT, OUTPUT_VERBOSITY_NORMAL, OUTPUT_VERBOSITY_VERBOSE, OUTPUT_VERBOSITY_EXTENDED,
        OUTPUT_VERBOSITY_RESULT_ONLY
    ]

    # list of all available Pygments styles (including ones loaded from plugins)
    # https://www.complang.tuwien.ac.at/doc/python-pygments/styles.html
    OUTPUT_STYLE = list(get_all_styles())

    WELCOME = """
    Welcome to {title} v{version}

    Press Ctrl-C to cancel the current command, and Ctrl-D to exit the shell.
    Type "help" to get help. Try TAB for auto-completion.
    """.format(title=style_crossbar('Crossbar.io Shell'), version=__version__)

    CONNECTED = """    Connection:

        url         : {url}
        authmethod  : {authmethod}
        realm       : {realm}
        authid      : {authid}
        authrole    : {authrole}
        session     : {session}
    """

    log = make_logger()

    def __init__(self):
        self.current_resource_type = None  # type: str
        self.current_resource = None
        self.session = None
        self._history = FileHistory('.cbsh-history')
        self._output_format = Application.OUTPUT_FORMAT_JSON_COLORED
        self._output_verbosity = Application.OUTPUT_VERBOSITY_NORMAL

        self._style = style_from_pygments_dict({
            Token.Toolbar: '#fce94f bg:#333333',

            # User input.
            # Token:          '#ff0066',

            # Prompt.
            # Token.Username: '#884444',
            # Token.At:       '#00aa00',
            # Token.Colon:    '#00aa00',
            # Token.Pound:    '#00aa00',
            # Token.Host:     '#000088 bg:#aaaaff',
            # Token.Path:     '#884444 underline',
        })

        self._output_style = 'fruity'

    @staticmethod
    def load_profile(dotdir=None, profile=None, yes_to_all=False, verbose=False):

        profile = profile or 'default'

        if not dotdir:
            if 'CROSSBAR_FABRIC_SUPERUSER' in os.environ:
                cbf_dir = os.path.abspath(os.path.dirname(os.environ['CROSSBAR_FABRIC_SUPERUSER']))
                if verbose:
                    click.echo('Using dotdir derived from CROSSBAR_FABRIC_SUPERUSER: {}'.format(style_ok(cbf_dir)))
            else:
                cbf_dir = os.path.abspath(os.path.expanduser('~/.crossbar'))
                if verbose:
                    click.echo('Using default dotdir: {}'.format(style_ok(cbf_dir)))
        else:
            cbf_dir = os.path.abspath(os.path.expanduser(dotdir))
            if verbose:
                click.echo('Using explicit dotdir: {}'.format(style_ok(cbf_dir)))

        if not os.path.isdir(cbf_dir):
            os.mkdir(cbf_dir)
            if verbose:
                click.echo('Created new local user directory: {}'.format(style_ok(cbf_dir)))

        config_path = os.path.join(cbf_dir, 'config.ini')
        if not os.path.isfile(config_path):
            with open(config_path, 'w') as f:
                url = _prompt_for_url(yes_to_all)
                f.write(_DEFAULT_CONFIG.format(url=url))
                if verbose:
                    click.echo('Created new local user configuration: {}'.format(style_ok(config_path)))
        else:
            if verbose:
                click.echo('Using existing local user configuration: {}'.format(style_ok(config_path)))

        config_obj = config.UserConfig(config_path)

        profile_obj = config_obj.profiles.get(profile, None)
        if not profile_obj:
            raise click.ClickException('no such profile: "{}"'.format(profile))
        else:
            if verbose:
                click.echo('Active user profile: {}'.format(style_ok(profile)))

        privkey_path = os.path.join(cbf_dir, profile_obj.privkey or '{}.priv'.format(profile))  # noqa: W503
        pubkey_path = os.path.join(cbf_dir, profile_obj.pubkey or '{}.pub'.format(profile))  # noqa: W503
        key_obj = userkey.UserKey(privkey_path, pubkey_path, yes_to_all=yes_to_all)

        return key_obj, profile_obj

    def set_output_format(self, output_format):
        """
        Set command output format.

        :param output_format: The verbosity to use.
        :type output_format: str
        """
        if output_format in Application.OUTPUT_FORMAT:
            self._output_format = output_format
        else:
            raise Exception('invalid value {} for output_format (not in {})'.format(
                output_format, Application.OUTPUT_FORMAT))

    def set_output_verbosity(self, output_verbosity):
        """
        Set command output verbosity.

        :param output_verbosity: The verbosity to use.
        :type output_verbosity: str
        """
        if output_verbosity in Application.OUTPUT_VERBOSITY:
            self._output_verbosity = output_verbosity
        else:
            raise Exception('invalid value {} for output_verbosity (not in {})'.format(
                output_verbosity, Application.OUTPUT_VERBOSITY))

    def set_output_style(self, output_style):
        """
        Set pygments syntax highlighting style ("theme") to be used for command result output.

        :param output_style: The style to use.
        :type output_style: str
        """
        if output_style in Application.OUTPUT_STYLE:
            self._output_style = output_style
        else:
            raise Exception('invalid value {} for output_style (not in {})'.format(output_style,
                                                                                   Application.OUTPUT_STYLE))

    def error(self, msg):
        click.echo()

    def format_selected(self):
        return '{} -> {}.\n'.format(self.current_resource_type, self.current_resource)

    def print_selected(self):
        click.echo(self.format_selected())

    def selected(self):
        return self.current_resource_type, self.current_resource

    def __str__(self):
        return 'Application(current_resource_type={}, current_resource={})'.format(self.current_resource_type,
                                                                                   self.current_resource)

    async def run_command(self, cmd):
        try:
            result = await cmd.run(self.session)
        except Exception as e:
            print(e)
        else:
            self._output_result(result)

    def _output_result(self, result):
        cmd_str = ' '.join(["crossbar", "shell"] + sys.argv[1:])
        if self._output_format in [Application.OUTPUT_FORMAT_JSON, Application.OUTPUT_FORMAT_JSON_COLORED]:

            json_str = json.dumps(result.result,
                                  separators=(', ', ': '),
                                  sort_keys=False,
                                  indent=4,
                                  ensure_ascii=False)

            if self._output_format == Application.OUTPUT_FORMAT_JSON_COLORED:
                console_str = highlight(json_str, lexers.JsonLexer(),
                                        formatters.Terminal256Formatter(style=self._output_style))
            else:
                console_str = json_str

        elif self._output_format in [Application.OUTPUT_FORMAT_YAML, Application.OUTPUT_FORMAT_YAML_COLORED]:

            yaml_str = yaml.safe_dump(result.result)

            if self._output_format == Application.OUTPUT_FORMAT_YAML_COLORED:
                console_str = highlight(yaml_str, lexers.YamlLexer(),
                                        formatters.Terminal256Formatter(style=self._output_style))
            else:
                console_str = yaml_str

        elif self._output_format == Application.OUTPUT_FORMAT_PLAIN:

            console_str = '{}'.format(result)

        else:
            # should not arrive here
            raise Exception('internal error: unprocessed value "{}" for output format'.format(self._output_format))

        # output command metadata (such as runtime)
        if self._output_verbosity == Application.OUTPUT_VERBOSITY_SILENT:
            pass
        else:
            # output result of command
            click.echo(console_str)

            if self._output_verbosity == Application.OUTPUT_VERBOSITY_RESULT_ONLY or self._output_format == Application.OUTPUT_FORMAT_PLAIN:
                pass
            elif self._output_verbosity == Application.OUTPUT_VERBOSITY_NORMAL:
                if result.duration:
                    click.echo(style_finished_line('Finished command in {} ms: {}'.format(result.duration, cmd_str)))
                else:
                    click.echo(style_finished_line('Finished command successfully: {}'.format(cmd_str)))
            elif self._output_verbosity == Application.OUTPUT_VERBOSITY_EXTENDED:
                if result.duration:
                    click.echo(
                        style_finished_line('Finished command in {} ms on {}: {}'.format(
                            result.duration, localnow(), cmd_str)))
                else:
                    click.echo(style_finished_line('Finished successfully on {}: {}'.format(localnow(), cmd_str)))
            else:
                # should not arrive here
                raise Exception('internal error')

    def _get_bottom_toolbar_tokens(self, cli):
        toolbar_str = ' Current resource path: {}'.format(self.format_selected())
        return [
            (Token.Toolbar, toolbar_str),
        ]

    def _get_prompt_tokens(self, cli):
        return [
            (Token.Username, 'john'),
            (Token.At, '@'),
            (Token.Host, 'localhost'),
            (Token.Colon, ':'),
            (Token.Path, '/user/john'),
            (Token.Pound, '# '),
        ]

    def run_context(self, ctx, command=None):

        # cfg contains the command lines options and arguments that
        # click collected for us
        cfg = ctx.obj
        cmd = ctx.command.name

        self.log.info('{klass}.run_context: running shell command "{cmd}"', klass=self.__class__.__name__, cmd=cmd)

        yes_to_all = cfg.yes_to_all if hasattr(cfg, 'yes_to_all') else False

        # if cmd not in ['auth', 'shell']:
        #    raise click.ClickException('"{}" command can only be run in shell'.format(cmd))

        if self._output_verbosity == Application.OUTPUT_VERBOSITY_VERBOSE:
            click.echo('Crossbar.io Shell: {}'.format(style_ok('v{}'.format(__version__))))

        # load user profile and key for given profile name
        key, profile = self.load_profile(dotdir=cfg.dotdir,
                                         profile=cfg.profile,
                                         yes_to_all=yes_to_all,
                                         verbose=(ctx.command.name == 'init'))

        if ctx.command.name == 'init':
            return

        # set the Fabric URL to connect to from the profile or default
        url = profile.url or 'wss://fabric.crossbario.com'

        # users always authenticate with the user_id from the key, which
        # filled from the email the user provided
        authid = key.user_id

        # the realm can be set from command line, env var, the profile
        # or can be None, which means the user will be joined to the global
        # Crossbar.io users realm ('com.crossbario.fabric')
        realm = cfg.realm or profile.realm or None

        # the authrole can be set from command line, env var, the profile
        # or can be None, in which case the role is chosen automatically
        # from the list of roles the user us authorized for
        authrole = cfg.role or profile.role or None

        # this will be fired when the ShellClient below actually has joined
        # the respective realm on Crossbar.io (either the global users
        # realm, or a management realm the user has a role on)
        done = txaio.create_future()

        url_is_secure, _, _, _, _, _ = parse_url(url)

        extra = {
            # these are forward on the actual client connection
            'authid': authid,
            'authrole': authrole,

            # these are native Python object and only used client-side
            'key': key.key,
            'done': done,
            'command': command,

            # WAMP-cryptosign authentication: TLS channel binding
            'channel_binding': 'tls-unique' if url_is_secure else None,
        }

        cert_options = None
        if profile.tls_hostname:
            self.log.info('Setting up TLS context (server CA/intermediate certificates, etc) from profile:')
            tls_config = {'hostname': profile.tls_hostname, 'ca_certificates': profile.tls_certificates}
            cert_options = _create_tls_client_context(tls_config, '.crossbar', self.log)

        # for the "auth" command, forward additional command line options
        if ctx.command.name == 'auth':
            # user provides authentication code to verify
            extra['activation_code'] = cfg.code

            # user requests sending of a new authentication code (while an old one is still pending)
            extra['request_new_activation_code'] = cfg.new_code

        # this is the WAMP ApplicationSession that connects the CLI to Crossbar.io
        self.session = client.ShellClient(ComponentConfig(realm, extra))

        runner = ApplicationRunner(url, realm, ssl=cert_options)

        if self._output_verbosity == Application.OUTPUT_VERBOSITY_VERBOSE:
            click.echo('Connecting to {} ..'.format(url))

        connect_done = runner.run(self.session, start_reactor=False)

        def on_connect_success(res):
            self.log.info('{klass}.on_connect_success(res={res})', klass=self.__class__.__name__, res=pformat(res))

        def on_connect_error(err):
            self.log.warn('{klass}.on_connect_error(err={err})', klass=self.__class__.__name__, err=err)

            if isinstance(err, Failure):
                err = err.value

            txaio.reject(done, err)

            # raise SystemExit(1)

        txaio.add_callbacks(connect_done, on_connect_success, on_connect_error)

        def on_success(res):
            self.log.info('{klass}.on_success(res={res})', klass=self.__class__.__name__, res=pformat(res))

            session_details, result = res

            if cmd == 'auth':

                self._print_welcome(url, session_details)

            elif cmd == 'shell':

                # click.clear()
                self._print_welcome(url, session_details)

                # FIXME:

                # prompt_kwargs = {
                #     'history': self._history,
                # }
                #
                # from crossbar.shell import repl
                #
                # shell_task = loop.create_task(
                #     repl.repl(
                #         ctx,
                #         get_bottom_toolbar_tokens=self._get_bottom_toolbar_tokens,
                #         # get_prompt_tokens=self._get_prompt_tokens,
                #         style=self._style,
                #         prompt_kwargs=prompt_kwargs))
                #
                # try:
                #     loop.run_until_complete(shell_task)
                # except Exception as e:
                #     print(e)

            else:
                if result:
                    self._output_result(result)

        def on_error(err):
            self.log.warn('{klass}.on_error(err={err})', klass=self.__class__.__name__, err=err)

            if isinstance(err, Failure):
                err = err.value

            if isinstance(err, ApplicationError):

                self.log.warn('{message} - {error}', message=err.args[0] if err.args else '', error=err.error)

                # some ApplicationErrors are actually signaling progress
                # in the authentication flow, some are real errors

                exit_code = None

                if err.error.startswith('fabric.auth-failed.'):
                    error = err.error.split('.')[2]
                    message = err.args[0]

                    if error == 'new-user-auth-code-sent':

                        click.echo('\nThanks for registering! {}'.format(message))
                        click.echo(
                            style_ok(
                                'Please check your inbox and run "crossbar shell auth --code <THE CODE YOU GOT BY EMAIL>.\n'
                            ))

                    elif error == 'registered-user-auth-code-sent':

                        click.echo('\nWelcome back! {}'.format(message))
                        click.echo(
                            style_ok(
                                'Please check your inbox and run "crossbar shell auth --code <THE CODE YOU GOT BY EMAIL>.\n'
                            ))

                    elif error == 'pending-activation':

                        click.echo()
                        click.echo(style_ok(message))
                        click.echo()
                        click.echo('Tip: to activate, run "crossbar shell auth --code <THE CODE YOU GOT BY EMAIL>"')
                        click.echo('Tip: you can request sending a new code with "crossbar shell auth --new-code"')
                        click.echo()

                    elif error == 'no-pending-activation':

                        exit_code = 1
                        click.echo()
                        click.echo(style_error('{} [{}]'.format(message, err.error)))
                        click.echo()

                    elif error == 'email-failure':

                        exit_code = 1
                        click.echo()
                        click.echo(style_error('{} [{}]'.format(message, err.error)))
                        click.echo()

                    elif error == 'invalid-activation-code':

                        exit_code = 1
                        click.echo()
                        click.echo(style_error('{} [{}]'.format(message, err.error)))
                        click.echo()

                    else:

                        exit_code = 1
                        click.echo(style_error('{}'.format(error)))
                        click.echo(style_error(message))

                elif err.error.startswith('crossbar.error.'):

                    error = err.error.split('.')[2]
                    message = err.args[0]

                    if error == 'invalid_configuration':

                        click.echo()
                        click.echo(style_error('{} [{}]'.format(message, err.error)))
                        click.echo()
                    else:

                        exit_code = 1
                        click.echo(style_error('{} [{}]'.format(message, err.error)))

                else:

                    click.echo(style_error('{}'.format(err)))
                    exit_code = 1

                if exit_code:
                    raise SystemExit(exit_code)

            else:
                click.echo(style_error('{}'.format(err)))
                raise SystemExit(1)

        txaio.add_callbacks(done, on_success, on_error)

        def doit(reactor):
            return done

        react(doit)

    def _print_welcome(self, url, session_details):
        click.echo(self.WELCOME)
        click.echo(
            self.CONNECTED.format(url=url,
                                  realm=style_crossbar(session_details.realm) if session_details else None,
                                  authmethod=session_details.authmethod if session_details else None,
                                  authid=style_crossbar(session_details.authid) if session_details else None,
                                  authrole=style_crossbar(session_details.authrole) if session_details else None,
                                  session=session_details.session if session_details else None))



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from datetime import datetime

from autobahn.util import utcstr


class RouterComponent(object):
    """
    A application component hosted and running inside a router worker.
    """
    def __init__(self, id, config, session):
        """

        :param id: The component ID within the router instance.
        :type id: str

        :param config: The component's configuration.
        :type config: dict

        :param session: The component application session.
        :type session: obj (instance of ApplicationSession)
        """
        self.id = id
        self.config = config
        self.session = session
        self.created = datetime.utcnow()

    def marshal(self):
        """
        Marshal object information for use with WAMP calls/events.
        """
        now = datetime.utcnow()
        return {
            'id': self.id,
            # 'started' is used by container-components; keeping it
            # for consistency in the public API
            'started': utcstr(self.created),
            'uptime': (now - self.created).total_seconds(),
            'config': self.config
        }


class RouterRealm(object):
    """
    A realm running in a router worker.
    """
    def __init__(self, controller, id, config, router=None, session=None):
        """

        :param controller: The controller this router is running under.
        :type controller: object

        :param id: The realm ID within the router worker, identifying the router.
        :type id: str

        :param config: The realm configuration.
        :type config: dict

        :param router: The router (within the router worker) serving the realm.
        :type router: :class:`crossbar.router.router.Router`

        :param session: The realm service session.
        :type session: :class:`crossbar.router.service.RouterServiceAgent`
        """
        # import here to dissolve circular dependency
        from crossbar.worker.rlink import RLinkManager

        self.controller = controller
        self.id = id
        self.config = config

        # this is filled later (after construction) when the router has been started
        self.router = router

        # this is filled later (after construction) when the router service agent session has been started
        self.session = session

        # router-realm links ("router-to-router connections")
        self.rlink_manager = RLinkManager(self, controller)

        self.created = datetime.utcnow()

        # Crossbar.io role run-time ID -> RouterRealmRole
        self.roles = {}

        # role WAMP name -> Crossbar.io role run-time ID
        self.role_to_id = {}

    def marshal(self):
        marshalled = {
            'id': self.id,
            'config': self.config,
            'created': utcstr(self.created),
            'roles': [self.roles[role].marshal() for role in self.roles if self.roles],
            'has_router': self.router is not None,
            'has_service_session': self.session is not None,
        }

        rlinks = []
        for link_id in self.rlink_manager.keys():
            rlinks.append(self.rlink_manager[link_id].marshal())

        marshalled['rlinks'] = rlinks

        return marshalled


class RouterRealmRole(object):
    """
    A role in a realm running in a router worker.
    """
    def __init__(self, id, config):
        """

        :param id: The role ID within the realm.
        :type id: str

        :param config: The role configuration.
        :type config: dict
        """
        self.id = id
        self.config = config

    def marshal(self):
        return {
            'id': self.id,
            'config': self.config,
        }



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import os
import uuid
import time
import binascii
import threading
from collections.abc import Mapping
from pprint import pformat
from pathlib import Path

import requests
import numpy as np
import web3
import multihash
from hexbytes import HexBytes
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet.threads import deferToThread

from txaio import make_logger, time_ns

import zlmdb
import cfxdb

from autobahn import wamp
from autobahn.wamp.exception import ApplicationError

from autobahn import xbr
from autobahn.xbr import unpack_uint256, pack_uint256, make_w3

from crossbar.node.worker import NativeWorkerProcess
from crossbar.worker.controller import WorkerController

from crossbar._util import hl, hlid, hlval, hlcontract
from crossbar.edge.worker.xbr import MarketMaker

__all__ = ('MarketplaceController', 'MarketplaceControllerProcess')


class MarketplaceControllerProcess(NativeWorkerProcess):

    TYPE = 'marketplace'
    LOGNAME = 'Marketplace'


class MarketplaceController(WorkerController):

    WORKER_TYPE = u'marketplace'
    WORKER_TITLE = u'Marketplace'

    STATUS_STARTING = 'starting'
    STATUS_RUNNING = 'running'
    STATUS_STOPPING = 'stopping'

    log = make_logger()

    def __init__(self, config=None, reactor=None, personality=None):
        # WorkerController derives of NativeProcess, which will set self._reactor
        WorkerController.__init__(self, config=config, reactor=reactor, personality=personality)

        worker_options_extra = dict(config.extra.extra)
        self._database_config = worker_options_extra['database']
        self._blockchain_config = worker_options_extra['blockchain']
        self._ipfs_files_directory = worker_options_extra.get('ipfs_files_directory', './.ipfs_files')

        # xbrmm worker status
        self._status = None

        # map of market makers by ID
        self._makers = {}
        self._maker_adr2id = {}

        # open xbrmm worker database, containing a replicate of xbr on-chain data (other than
        # channels, which are market specific and stored in the market maker database of the maker of that market)
        self._dbpath = os.path.abspath(
            self._database_config.get('dbpath', './.xbrmm-{}-db'.format(config.extra.worker)))
        self._db = zlmdb.Database(dbpath=self._dbpath,
                                  maxsize=self._database_config.get('maxsize', 2**30),
                                  readonly=False,
                                  sync=True)
        self._db.__enter__()

        # generic database object metadata
        self._meta = cfxdb.meta.Schema.attach(self._db)

        # xbr database schema
        self._xbr = cfxdb.xbr.Schema.attach(self._db)

        # xbr market maker schema
        self._xbrmm = cfxdb.xbrmm.Schema.attach(self._db)

        # event object too coordinate exit of blockchain monitor background check
        self._run_monitor = None

        # blockchain gateway configuration
        self._bc_gw_config = self._blockchain_config['gateway']
        self.log.info('Initializing Web3 from blockchain gateway configuration\n\n{gateway}\n',
                      gateway=pformat(self._bc_gw_config))
        self._w3 = make_w3(self._bc_gw_config)
        xbr.setProvider(self._w3)

        self._chain_id = self._blockchain_config.get('chain_id', 1)
        self.log.info('Using chain ID {chain_id}', chain_id=hlid(self._chain_id))

        # To be initiated once cbdir variable gets available
        self._ipfs_files_dir = os.path.join(config.extra.cbdir, self._ipfs_files_directory)

    @inlineCallbacks
    def onJoin(self, details, publish_ready=True):
        self.log.info(
            'XBR Markets Worker starting (realm={realm}, prefix="{prefix}", session={session}, authid={authid}, authrole={authrole})',
            realm=hlid(details.realm),
            prefix=hlid(self._uri_prefix),
            session=hlid(details.session),
            authid=hlid(details.authid),
            authrole=hlid(details.authrole))
        self.log.info('XBR Markets Worker configuration:\n\n{config}', config=pformat(self.config.extra))

        self._status = self.STATUS_STARTING

        yield WorkerController.onJoin(self, details, publish_ready=False)

        # any special session setup for the market maker goes here ..

        # start blockchain monitor is running on a background thread which exits once this gets False
        self._run_monitor = threading.Event()
        self._stop_monitor = False

        # FIXME: check self.xbr.blocks for latest block already processed
        # initially begin scanning the blockchain with this block, and subsequently scan from the last
        # processed and locally persisted block record in the database
        if 'from_block' in self._blockchain_config:
            scan_from_block = self._blockchain_config['from_block']
            self.log.info('Initial scanning of blockchain beginning with block {scan_from_block} from configuration',
                          scan_from_block=scan_from_block)
        else:
            scan_from_block = 1
            self.log.info('Initial scanning of blockchain from block 1 (!)')

        # monitor/pull blockchain from a background thread
        self._monitor_blockchain_thread = self._reactor.callInThread(self._monitor_blockchain, self._bc_gw_config,
                                                                     scan_from_block)
        self._status = self.STATUS_RUNNING

        yield self.publish_ready()

        self.log.info('XBR Markets Worker ready!')

    @inlineCallbacks
    def onLeave(self, details):
        """

        :param details:
        :return:
        """
        self.log.info('XBR Markets Worker shutting down ({market_cnt} markets to shutdown) ..',
                      market_cnt=len(self._makers))

        self._status = self.STATUS_STOPPING

        # shutdown each market ..
        makers = list(self._makers.values())
        for maker in makers:
            yield maker.stop()
            self.log.info('Market Maker "{maker_id}" stopped.', maker_id=hlid(maker._id))

        self._stop_monitor = True

        # stop blockchain monitoring background thread, possibly waking up
        # the thread from some blocking activity (like sleeping in a syscall)
        if not self._run_monitor.is_set():
            self._run_monitor.set()

        # make sure to join the background thread, so this main thread running
        # in the process exits - and thus the whole process
        if self._monitor_blockchain_thread:
            self._monitor_blockchain_thread.join()

        # disconnect from router
        self.disconnect()

        self._status = None

        self.log.info('XBR Markets Worker shutdown complete!')

    def _trigger_monitor_blockchain(self):
        self.log.info('Trigger (explicitly) the background blockchain monitor ..')
        self._run_monitor.set()
        self._run_monitor.clear()

    def _download_ipfs_file(self, file_hash):
        if not os.path.exists(self._ipfs_files_dir):
            Path(self._ipfs_files_dir).mkdir()

        file_path = os.path.join(self._ipfs_files_dir, file_hash)
        if os.path.exists(file_path):
            # If file exists in storage but not in db, then just add
            # a db entry.
            with self._db.begin(write=True) as txn:
                ipfs_file = self._xbrmm.ipfs_files[txn, file_hash]
                if not ipfs_file:
                    ipfs_file = cfxdb.xbrmm.IPFSFile()
                    ipfs_file.file_hash = file_hash
                    self._xbrmm.ipfs_files[txn, file_hash] = ipfs_file
        else:
            ipfs_file = cfxdb.xbrmm.IPFSFile()
            ipfs_file.file_hash = file_hash
            path = 'https://ipfs.infura.io:5001/api/v0/cat?arg={}'.format(file_hash)
            response = requests.get(path)
            if response.status_code == 200:
                with open(file_path, 'w') as file:
                    file.write(response.text)
                with self._db.begin(write=True) as txn:
                    self._xbrmm.ipfs_files[txn, file_hash] = ipfs_file
            else:
                ipfs_file.retries = ipfs_file.retries + 1
                ipfs_file.errored_at = np.datetime64(time_ns(), 'ns')
                with self._db.begin(write=True) as txn:
                    self._xbrmm.ipfs_files[txn, file_hash] = ipfs_file

    def _monitor_blockchain(self, gateway_config, scan_from_block, period=300):
        """

        :param gateway_config:
        :param scan_from_block:
        :param period:
        :return:
        """
        w3 = make_w3(gateway_config)
        initial_delay = 2

        self.log.info(
            'Start monitoring of blockchain ({blockchain_type}) on thread {thread_id} in {initial_delay} seconds, iterating every {period} seconds  ..',
            blockchain_type=str(self._bc_gw_config['type']),
            initial_delay=hlval(initial_delay),
            period=hlval(period),
            thread_id=hlval(int(threading.get_ident())))

        # using normal blocking call here, as we are running on a background thread!
        time.sleep(initial_delay)

        def _process_Token_Transfer(transactionHash, blockHash, args):
            # event Transfer(address indexed from, address indexed to, uint256 value);
            self.log.info(
                '{event}: processing event (tx_hash={tx_hash}, block_hash={block_hash}) - {value} XBR token transferred (on-chain) from {_from} to {_to})',
                event=hlcontract('XBRToken.Transfer'),
                tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()),
                block_hash=hlid('0x' + binascii.b2a_hex(blockHash).decode()),
                value=hlval(int(args.value / 10**18)),
                _from=hlval(args['from']),
                _to=hlval(args.to))

            stored = False
            with self._db.begin(write=True) as txn:

                transactionHash = bytes(transactionHash)

                token_transfer = self._xbr.token_transfers[txn, transactionHash]
                if token_transfer:
                    self.log.warn('{contract}(tx_hash={tx_hash}) record already stored in database.',
                                  contract=hlcontract('TokenTransfer'),
                                  tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()))
                else:
                    token_transfer = cfxdb.xbr.token.TokenTransfer()

                    token_transfer.tx_hash = transactionHash
                    token_transfer.block_hash = bytes(blockHash)
                    token_transfer.from_address = bytes(HexBytes(args['from']))
                    token_transfer.to_address = bytes(HexBytes(args.to))
                    token_transfer.value = args.value

                    self._xbr.token_transfers[txn, transactionHash] = token_transfer
                    stored = True

            if stored:
                self.log.info('new {contract}(tx_hash={tx_hash}) record stored database!',
                              contract=hlcontract('TokenTransfer'),
                              tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()))

        def _process_Token_Approval(transactionHash, blockHash, args):
            # event Approval(address indexed from, address indexed to, uint256 value);
            self.log.info(
                '{event}: processing event (tx_hash={tx_hash}, block_hash={block_hash}) - {value} XBR token approved (on-chain) from owner {owner} to spender {spender})',
                event=hlcontract('XBRToken.Approval'),
                tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()),
                block_hash=hlid('0x' + binascii.b2a_hex(blockHash).decode()),
                value=hlval(int(args.value / 10**18)),
                owner=hlval(args.owner),
                spender=hlval(args.spender))

            stored = False
            with self._db.begin(write=True) as txn:

                transactionHash = bytes(transactionHash)

                token_approval = self._xbr.token_approvals[txn, transactionHash]
                if token_approval:
                    self.log.warn('{contract}(tx_hash={tx_hash}) record already stored in database.',
                                  contract=hlcontract('TokenApproval'),
                                  tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()))
                else:
                    token_approval = cfxdb.xbr.token.TokenApproval()

                    token_approval.tx_hash = transactionHash
                    token_approval.block_hash = bytes(blockHash)
                    token_approval.owner_address = bytes(HexBytes(args.owner))
                    token_approval.spender_address = bytes(HexBytes(args.spender))
                    token_approval.value = args.value

                    self._xbr.token_approvals[txn, transactionHash] = token_approval
                    stored = True

            if stored:
                self.log.info('new {contract}(tx_hash={tx_hash}) record stored database!',
                              contract=hlcontract('TokenApproval'),
                              tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()))

        def _process_Network_MemberRegistered(transactionHash, blockHash, args):
            #     /// Event emitted when a new member joined the XBR Network.
            #     event MemberCreated (address indexed member, string eula, string profile, MemberLevel level);
            self.log.info(
                '{event}: processing event (tx_hash={tx_hash}, block_hash={block_hash}) - XBR member created at address {address})',
                event=hlcontract('XBRNetwork.MemberCreated'),
                tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()),
                block_hash=hlid('0x' + binascii.b2a_hex(blockHash).decode()),
                address=hlid(args.member))

            member_adr = bytes(HexBytes(args.member))

            if args.eula:
                h = multihash.decode(multihash.from_b58_string(args.eula))
                if h.name != 'sha2-256':
                    self.log.warn(
                        'WARNING: XBRNetwork.MemberCreated - eula "{eula}" is not an IPFS (sha2-256) b58-encoded multihash',
                        eula=hlval(args.eula))

            if args.profile:
                h = multihash.decode(multihash.from_b58_string(args.profile))
                if h.name != 'sha2-256':
                    self.log.warn(
                        'WARNING: XBRNetwork.MemberCreated - profile "{profile}" is not an IPFS (sha2-256) b58-encoded multihash',
                        eula=hlval(args.profile))

            stored = False
            with self._db.begin(write=True) as txn:

                member = self._xbr.members[txn, member_adr]
                if member:
                    self.log.warn('{contract}(tx_hash={tx_hash}) record already stored in database.',
                                  contract=hlcontract('TokenApproval'),
                                  tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()))
                else:
                    member = cfxdb.xbr.member.Member()
                    member.address = member_adr
                    member.timestamp = np.datetime64(time_ns(), 'ns')
                    member.registered = args.registered
                    member.eula = args.eula
                    member.profile = args.profile
                    member.level = args.level

                    self._xbr.members[txn, member_adr] = member
                    stored = True

            if stored:
                self.log.info('new {contract}(member_adr={member_adr}) record stored database!',
                              contract=hlcontract('MemberCreated'),
                              member_adr=hlid('0x' + binascii.b2a_hex(member_adr).decode()))

        def _process_Network_MemberRetired(transactionHash, blockHash, args):
            #     /// Event emitted when a member leaves the XBR Network.
            #     event MemberRetired (address member);
            self.log.warn('_process_Network_MemberRetired not implemented')

        def _process_Market_MarketCreated(transactionHash, blockHash, args):
            #     /// Event emitted when a new market was created.
            #     event MarketCreated (bytes16 indexed marketId, uint32 marketSeq, address owner, string terms, string meta,
            #         address maker, uint256 providerSecurity, uint256 consumerSecurity, uint256 marketFee);
            self.log.info(
                '{event}: processing event (tx_hash={tx_hash}, block_hash={block_hash}) - XBR market created with ID {market_id})',
                event=hlcontract('XBRMarket.MarketCreated'),
                tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()),
                block_hash=hlid('0x' + binascii.b2a_hex(blockHash).decode()),
                market_id=hlid(uuid.UUID(bytes=args.marketId)))

            market_id = uuid.UUID(bytes=args.marketId)

            if args.terms:
                h = multihash.decode(multihash.from_b58_string(args.terms))
                if h.name != 'sha2-256':
                    self.log.warn(
                        'WARNING: XBRMarket.MarketCreated - terms "{terms}" is not an IPFS (sha2-256) b58-encoded multihash',
                        terms=hlval(args.terms))

            if args.meta:
                h = multihash.decode(multihash.from_b58_string(args.meta))
                if h.name != 'sha2-256':
                    self.log.warn(
                        'WARNING: XBRMarket.MarketCreated - meta "{meta}" is not an IPFS (sha2-256) b58-encoded multihash',
                        meta=hlval(args.meta))

            stored = False
            with self._db.begin(write=True) as txn:

                market = self._xbr.markets[txn, market_id]
                if market:
                    self.log.warn('{contract}(tx_hash={tx_hash}) record already stored in database.',
                                  contract=hlcontract('MarketCreated'),
                                  tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()))
                else:
                    market = cfxdb.xbr.market.Market()
                    market.market = market_id
                    market.timestamp = np.datetime64(time_ns(), 'ns')

                    # FIXME
                    # market.created = args.created

                    market.seq = args.marketSeq
                    market.owner = bytes(HexBytes(args.owner))
                    market.terms = args.terms
                    market.meta = args.meta
                    market.maker = bytes(HexBytes(args.maker))
                    market.provider_security = args.providerSecurity
                    market.consumer_security = args.consumerSecurity
                    market.market_fee = args.marketFee

                    self._xbr.markets[txn, market_id] = market
                    stored = True

            if stored:
                self.log.info('new {contract}(market_id={market_id}) record stored database!',
                              contract=hlcontract('MarketCreated'),
                              market_id=hlid(market_id))

        def _process_Market_MarketUpdated(transactionHash, blockHash, args):
            #     /// Event emitted when a market was updated.
            #     event MarketUpdated (bytes16 indexed marketId, uint32 marketSeq, address owner, string terms, string meta,
            #         address maker, uint256 providerSecurity, uint256 consumerSecurity, uint256 marketFee);
            self.log.warn('_process_Market_MarketUpdated not implemented')

        def _process_Market_MarketClosed(transactionHash, blockHash, args):
            #     /// Event emitted when a market was closed.
            #     event MarketClosed (bytes16 indexed marketId);
            self.log.warn('_process_Market_MarketClosed not implemented')

        def _process_Market_ActorJoined(transactionHash, blockHash, args):
            # Event emitted when a new actor joined a market.
            # event ActorJoined (bytes16 indexed marketId, address actor, uint8 actorType, uint joined, uint256 security, string meta);
            self.log.info(
                '{event}: processing event (tx_hash={tx_hash}, block_hash={block_hash}) - XBR market actor {actor} joined market {market_id})',
                event=hlcontract('XBRMarket.ActorJoined'),
                tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()),
                block_hash=hlid('0x' + binascii.b2a_hex(blockHash).decode()),
                actor=hlid(args.actor),
                market_id=hlid(uuid.UUID(bytes=args.marketId)))

            market_id = uuid.UUID(bytes=args.marketId)
            actor_adr = bytes(HexBytes(args.actor))
            actor_type = int(args.actorType)

            if args.meta:
                h = multihash.decode(multihash.from_b58_string(args.meta))
                if h.name != 'sha2-256':
                    self.log.warn(
                        'WARNING: XBRMarket.MarketCreated - meta "{meta}" is not an IPFS (sha2-256) b58-encoded multihash',
                        terms=hlval(args.meta))

            stored = False
            with self._db.begin(write=True) as txn:

                actor = self._xbr.actors[txn, (market_id, actor_adr, actor_type)]
                if actor:
                    self.log.warn('{contract}(tx_hash={tx_hash}) record already stored in database.',
                                  contract=hlcontract('MarketCreated'),
                                  tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()))
                else:
                    actor = cfxdb.xbr.actor.Actor()
                    actor.timestamp = np.datetime64(time_ns(), 'ns')
                    actor.market = market_id
                    actor.actor = actor_adr
                    actor.actor_type = actor_type

                    actor.joined = args.joined
                    actor.security = args.security
                    actor.meta = args.meta

                    self._xbr.actors[txn, (market_id, actor_adr, actor_type)] = actor
                    stored = True

            if stored:
                self.log.info(
                    'new {contract}(market_id={market_id}, actor_adr={actor_adr}, actor_type={actor_type}) record stored database!',
                    contract=hlcontract('ActorJoined'),
                    market_id=hlid(market_id),
                    actor_adr=hlid('0x' + binascii.b2a_hex(actor_adr).decode()),
                    actor_type=hlid(actor_type))

        def _process_Market_ActorLeft(transactionHash, blockHash, args):
            self.log.warn('_process_Market_ActorLeft not implemented')

        def _process_Market_ConsentSet(transactionHash, blockHash, args):
            # Event emitted when a consent is set
            # emit ConsentSet(member, updated, marketId, delegate, delegateType,
            #                 apiCatalog, consent, servicePrefix);
            self.log.info('{event}: processing event (tx_hash={tx_hash}, block_hash={block_hash}) ..',
                          event=hlcontract('XBRMarket.ConsentSet'),
                          tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()),
                          block_hash=hlid('0x' + binascii.b2a_hex(blockHash).decode()))

            catalog_oid = uuid.UUID(bytes=args.apiCatalog)
            member = uuid.UUID(bytes=args.member)
            delegate = args.delegate
            delegate_type = args.delegateType
            market_oid = uuid.UUID(bytes=args.marketId)
            with self._db.begin(write=True) as txn:
                consent = self._xbr.consents[txn, (catalog_oid, member, delegate, delegate_type, market_oid)]
                consent.synced = True

        def _process_Catalog_CatalogCreated(transactionHash, blockHash, args):
            # Event emitted when a new API catalog is created
            # emit CatalogCreated(catalogId, created, catalogSeq, owner, terms, meta);
            self.log.info('{event}: processing event (tx_hash={tx_hash}, block_hash={block_hash}) ..',
                          event=hlcontract('XBRCatalog.CatalogCreated'),
                          tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()),
                          block_hash=hlid('0x' + binascii.b2a_hex(blockHash).decode()))

            catalog_oid = uuid.UUID(bytes=args.catalogId)
            owner = bytes(HexBytes(args.owner))
            created = np.datetime64(time_ns(), 'ns')
            with self._db.begin(write=True) as txn:
                catalog = cfxdb.xbr.catalog.Catalog()
                catalog.oid = catalog_oid
                catalog.timestamp = created
                catalog.seq = args.catalogSeq
                catalog.owner = owner
                catalog.terms = args.terms
                catalog.meta = args.meta
                self._xbr.catalogs[txn, catalog_oid] = catalog

            deferToThread(self._download_ipfs_file, args.meta)

        def _process_Catalog_ApiPublished(transactionHash, blockHash, args):
            self.log.warn('_process_Catalog_ApiPublished not implemented')

        def _process_Channel_Opened(transactionHash, blockHash, args):
            # Event emitted when a new XBR data market has opened.
            # event Opened(XBRTypes.ChannelType ctype, bytes16 indexed marketId, bytes16 indexed channelId,
            #              address actor, address delegate, address marketmaker, address recipient,
            #              uint256 amount, bytes signature);
            self.log.info('{event}: processing event (tx_hash={tx_hash}, block_hash={block_hash}) ..',
                          event=hlcontract('XBRChannel.Opened'),
                          tx_hash=hlid('0x' + binascii.b2a_hex(transactionHash).decode()),
                          block_hash=hlid('0x' + binascii.b2a_hex(blockHash).decode()))

            channel_oid = uuid.UUID(bytes=args.channelId)
            marketmaker_adr = bytes(HexBytes(args.marketmaker))
            marketmaker_adr_str = web3.Web3.toChecksumAddress(marketmaker_adr)

            # we only persist data for xbr markets operated by one of the market makers we run in this worker
            if marketmaker_adr_str not in self._maker_adr2id or self._maker_adr2id[
                    marketmaker_adr_str] not in self._makers:
                self.log.info(
                    '{event}: skipping channel (channel {channel_oid} in market with market maker address {marketmaker_adr} is in for any market in this markets worker)',
                    event=hlcontract('XBRChannel.Opened'),
                    channel_oid=hlid(channel_oid),
                    marketmaker_adr=hlid(marketmaker_adr_str))
                return

            # prepare new channel data
            channel_type = int(args.ctype)
            market_oid = uuid.UUID(bytes=args.marketId)
            actor_adr = bytes(HexBytes(args.actor))
            delegate_adr = bytes(HexBytes(args.delegate))
            recipient_adr = bytes(HexBytes(args.recipient))
            amount = int(args.amount)
            # FIXME
            # signature = bytes(HexBytes(args.signature))
            # FIXME
            # member_oid = uuid.UUID()

            # get the market maker by address
            maker = self._makers[self._maker_adr2id[marketmaker_adr_str]]

            # the market maker specific embedded database and schema
            db = maker.db
            xbrmm = maker.schema

            # depending on channel type, different database schema classes and tables are used
            if channel_type == cfxdb.xbrmm.ChannelType.PAYMENT:
                channel = cfxdb.xbrmm.PaymentChannel()
                channels = xbrmm.payment_channels
                channels_by_delegate = xbrmm.idx_payment_channel_by_delegate
                balance = cfxdb.xbrmm.PaymentChannelBalance()
                balances = xbrmm.payment_balances
            elif channel_type == cfxdb.xbrmm.ChannelType.PAYING:
                channel = cfxdb.xbrmm.PayingChannel()
                channels = xbrmm.paying_channels
                channels_by_delegate = xbrmm.idx_paying_channel_by_delegate
                balance = cfxdb.xbrmm.PayingChannelBalance()
                balances = xbrmm.paying_balances
            else:
                assert False, 'should not arrive here'

            # fill in information for newly replicated channel
            channel.market_oid = market_oid
            # FIXME
            # channel.member_oid = member_oid
            channel.channel_oid = channel_oid
            channel.timestamp = np.datetime64(time_ns(), 'ns')
            # channel.open_at = None

            # FIXME: should read that from even args after deployment of
            # https://github.com/crossbario/xbr-protocol/pull/138
            channel.seq = 1
            channel.channel_type = channel_type
            channel.marketmaker = marketmaker_adr
            channel.actor = actor_adr
            channel.delegate = delegate_adr
            channel.recipient = recipient_adr
            channel.amount = amount

            # FIXME
            channel.timeout = 0
            channel.state = cfxdb.xbrmm.ChannelState.OPEN
            # FIXME
            # channel.open_sig = signature

            # create an off-chain balance record for the channel with remaining == initial amount
            balance.remaining = channel.amount
            # FIXME: should read that from even args after deployment of
            # https://github.com/crossbario/xbr-protocol/pull/138
            balance.seq = 1

            # now store the new channel and balance in the database
            stored = False
            cnt_channels_before = 0
            cnt_channels_by_delegate_before = 0
            cnt_channels_after = 0
            cnt_channels_by_delegate_after = 0
            with db.begin(write=True) as txn:
                if channels[txn, channel_oid]:
                    self.log.warn('{event}: channel already stored in database [channel_oid={channel_oid}]',
                                  event=hlcontract('XBRChannel.Opened'),
                                  channel_oid=hlid(channel_oid))
                else:
                    cnt_channels_before = channels.count(txn)
                    cnt_channels_by_delegate_before = channels_by_delegate.count(txn)

                    # store the channel along with the off-chain balance
                    channels[txn, channel_oid] = channel
                    balances[txn, channel_oid] = balance
                    stored = True

                    cnt_channels_after = channels.count(txn)
                    cnt_channels_by_delegate_after = channels_by_delegate.count(txn)

            self.log.info(
                '{event} DB result: stored={stored}, cnt_channels_before={cnt_channels_before}, cnt_channels_by_delegate_before={cnt_channels_by_delegate_before}, cnt_channels_after={cnt_channels_after}, cnt_channels_by_delegate_after={cnt_channels_by_delegate_after}',
                event=hlcontract('XBRChannel.Opened'),
                stored=hlval(stored),
                cnt_channels_before=hlval(cnt_channels_before),
                cnt_channels_by_delegate_before=hlval(cnt_channels_by_delegate_before),
                cnt_channels_after=hlval(cnt_channels_after),
                cnt_channels_by_delegate_after=hlval(cnt_channels_by_delegate_after))
            if stored:
                # FIXME: publish WAMP event
                self.log.info(
                    '{event}: new channel stored in database [actor_adr={actor_adr}, channel_type={channel_type}, market_oid={market_oid}, member_oid={member_oid}, channel_oid={channel_oid}]',
                    event=hlcontract('XBRChannel.Opened'),
                    market_oid=hlid(market_oid),
                    # FIXME
                    member_oid=hlid(None),
                    channel_oid=hlid(channel_oid),
                    actor_adr=hlid('0x' + binascii.b2a_hex(actor_adr).decode()),
                    channel_type=hlid(channel_type))

        def _process_Channel_Closing(transactionHash, blockHash, args):
            self.log.warn('_process_Channel_Closing not implemented')

        def _process_Channel_Closed(transactionHash, blockHash, args):
            self.log.warn('_process_Channel_Closed not implemented')

        # map XBR contract log event to event processing function
        Events = [
            (xbr.xbrtoken.events.Transfer, _process_Token_Transfer),
            (xbr.xbrtoken.events.Approval, _process_Token_Approval),
            (xbr.xbrnetwork.events.MemberRegistered, _process_Network_MemberRegistered),
            (xbr.xbrnetwork.events.MemberRetired, _process_Network_MemberRetired),
            (xbr.xbrmarket.events.MarketCreated, _process_Market_MarketCreated),
            (xbr.xbrmarket.events.MarketUpdated, _process_Market_MarketUpdated),
            (xbr.xbrmarket.events.MarketClosed, _process_Market_MarketClosed),
            (xbr.xbrmarket.events.ActorJoined, _process_Market_ActorJoined),
            (xbr.xbrmarket.events.ActorLeft, _process_Market_ActorLeft),
            (xbr.xbrmarket.events.ConsentSet, _process_Market_ConsentSet),
            (xbr.xbrcatalog.events.CatalogCreated, _process_Catalog_CatalogCreated),
            (xbr.xbrcatalog.events.ApiPublished, _process_Catalog_ApiPublished),
            (xbr.xbrchannel.events.Opened, _process_Channel_Opened),
            (xbr.xbrchannel.events.Closing, _process_Channel_Closing),
            (xbr.xbrchannel.events.Closed, _process_Channel_Closed),
        ]

        # determine the block number, starting from which we scan the blockchain for XBR events
        current = w3.eth.getBlock('latest')
        last_processed = scan_from_block - 1
        with self._db.begin() as txn:
            for block_number in self._xbr.blocks.select(txn, return_values=False, reverse=True, limit=1):
                last_processed = unpack_uint256(block_number)
        if last_processed > current.number:
            raise ApplicationError(
                'wamp.error.invalid_argument',
                'last processed block number {} (or configured "scan_from" block number) is larger than then current block number {}'
                .format(last_processed, current.number))
        else:
            self.log.info(
                'Start scanning blockchain: current block is {current_block}, last processed is {last_processed} ..',
                current_block=hlval(current.number),
                last_processed=hlval(last_processed + 1))

        iteration = 1

        while not self._stop_monitor and not self._run_monitor.is_set():
            # current last block
            current = w3.eth.getBlock('latest')

            # track number of blocks processed
            cnt_blocks_success = 0
            cnt_blocks_error = 0
            cnt_xbr_events = 0

            # synchronize on-change changes locally by processing blockchain events
            if last_processed < current.number:
                while last_processed < current.number:
                    last_processed += 1
                    try:
                        self.log.info('Now processing blockchain block {last_processed} ..',
                                      last_processed=hlval(last_processed))
                        cnt_xbr_events += self._process_block(w3, last_processed, Events)
                    except:
                        self.log.failure()
                        cnt_blocks_error += 1
                    else:
                        cnt_blocks_success += 1

                self.log.info(
                    'Monitor blockchain iteration {iteration} completed: new block processed (last_processed={last_processed}, thread_id={thread_id}, period={period}, cnt_xbr_events={cnt_xbr_events}, cnt_blocks_success={cnt_blocks_success}, cnt_blocks_error={cnt_blocks_error})',
                    iteration=hlval(iteration),
                    last_processed=hlval(last_processed),
                    thread_id=hlval(int(threading.get_ident())),
                    period=hlval(period),
                    cnt_xbr_events=hlval(cnt_xbr_events, color='green') if cnt_xbr_events else hlval(cnt_xbr_events),
                    cnt_blocks_success=hlval(cnt_blocks_success, color='green')
                    if cnt_xbr_events else hlval(cnt_blocks_success),
                    cnt_blocks_error=hlval(cnt_blocks_error, color='red')
                    if cnt_blocks_error else hlval(cnt_blocks_error))
            else:
                self.log.info(
                    'Monitor blockchain iteration {iteration} completed: no new blocks found (last_processed={last_processed}, thread_id={thread_id}, period={period})',
                    iteration=hlval(iteration),
                    last_processed=hlval(last_processed),
                    thread_id=hlval(int(threading.get_ident())),
                    period=hlval(period))

            # sleep (using normal blocking call here, as we are running on a background thread!)
            self._run_monitor.wait(period)

            iteration += 1

    def _process_block(self, w3, block_number, Events):
        """

        :param w3:
        :param block_number:
        :param Events:
        :return:
        """
        cnt = 0
        # filter by block, and XBR contract addresses
        # FIXME: potentially add filters for global data or market specific data for the markets started in this worker
        filter_params = {
            'address': [
                xbr.xbrtoken.address, xbr.xbrnetwork.address, xbr.xbrcatalog.address, xbr.xbrmarket.address,
                xbr.xbrchannel.address
            ],
            'fromBlock':
            block_number,
            'toBlock':
            block_number,
        }
        result = w3.eth.getLogs(filter_params)
        if result:
            for evt in result:
                receipt = w3.eth.getTransactionReceipt(evt['transactionHash'])
                for Event, handler in Events:
                    # FIXME: MismatchedABI pops up .. we silence this with errors=web3.logs.DISCARD
                    if hasattr(web3, 'logs') and web3.logs:
                        all_res = Event().processReceipt(receipt, errors=web3.logs.DISCARD)
                    else:
                        all_res = Event().processReceipt(receipt)
                    for res in all_res:
                        self.log.info('{handler} processing block {block_number} / txn {txn} with args {args}',
                                      handler=hl(handler.__name__),
                                      block_number=hlid(block_number),
                                      txn=hlid('0x' + binascii.b2a_hex(evt['transactionHash']).decode()),
                                      args=hlval(res.args))
                        handler(res.transactionHash, res.blockHash, res.args)
                        cnt += 1

        with self._db.begin(write=True) as txn:
            block = cfxdb.xbr.block.Block()
            block.timestamp = np.datetime64(time_ns(), 'ns')
            block.block_number = block_number
            # FIXME
            # block.block_hash = bytes()
            block.cnt_events = cnt
            self._xbr.blocks[txn, pack_uint256(block_number)] = block

        if cnt:
            self.log.info('Processed blockchain block {block_number}: processed {cnt} XBR events.',
                          block_number=hlid(block_number),
                          cnt=hlid(cnt))
        else:
            self.log.info('Processed blockchain block {block_number}: no XBR events found!',
                          block_number=hlid(block_number))

        return cnt

    def _get_transaction_receipt(self, transaction: bytes):
        """

        :param transaction:
        :return:
        """
        # get the full transaction receipt given the transaction hash
        receipt = self._w3.eth.getTransactionReceipt(transaction)
        return receipt

    def _get_gas_price(self):
        """

        :return:
        """
        # FIXME: read from eth gas station
        return self._w3.toWei('10', 'gwei')

    def _get_balances(self, wallet_adr):
        """

        :param wallet_adr:
        :return:
        """
        eth_balance = self._w3.eth.getBalance(wallet_adr)
        xbr_balance = xbr.xbrtoken.functions.balanceOf(wallet_adr).call()
        return eth_balance, xbr_balance

    @inlineCallbacks
    @wamp.register(None)
    def start_market_maker(self, maker_id, config, details=None):
        """
        Starts a XBR Market Maker providing services in a specific XBR market.
        """
        if type(maker_id) != str:
            emsg = 'maker_id has invalid type {}'.format(type(maker_id))
            raise ApplicationError('wamp.error.invalid_argument', emsg)

        if not isinstance(config, Mapping):
            emsg = 'maker_id has invalid type {}'.format(type(config))
            raise ApplicationError('wamp.error.invalid_argument', emsg)

        if maker_id in self._makers:
            emsg = 'could not start market maker: a market maker with ID "{}" is already running (or starting)'.format(
                maker_id)
            raise ApplicationError('crossbar.error.already_running', emsg)

        self.personality.check_market_maker(self.personality, config)

        self.log.info('XBR Market Maker "{maker_id}" starting with config:\n{config}',
                      maker_id=hlid(maker_id),
                      config=pformat(config))

        maker = MarketMaker(self, maker_id, config, self._db, self._ipfs_files_dir)
        self._makers[maker_id] = maker
        self._maker_adr2id[maker.address] = maker_id

        yield maker.start()

        status = yield maker.status()
        self.log.info('{msg}: {accounts} local accounts, current block number is {current_block_no}',
                      msg=hl('Blockchain status', color='green', bold=True),
                      current_block_no=hlid(status['current_block_no']),
                      accounts=hlid(len(status['accounts'])))

        started = {
            'id': maker_id,
            'address': maker.address,
        }
        self.publish(u'{}.on_maker_started'.format(self._uri_prefix), started)

        self.log.info(
            'XBR Market Maker "{maker_id}" (address {maker_adr}) started. Now running {maker_cnt} market makers in total in this worker component.',
            maker_id=maker_id,
            maker_adr=maker.address,
            maker_cnt=len(self._makers))

        returnValue(started)

    @inlineCallbacks
    @wamp.register(None)
    def stop_market_maker(self, maker_id, details=None):
        if type(maker_id) != str:
            emsg = 'maker_id has invalid type {}'.format(type(maker_id))
            raise ApplicationError('wamp.error.invalid_argument', emsg)

        if maker_id not in self._makers:
            emsg = 'could not stop market maker: no market maker with ID "{}" is currently running'.format(maker_id)
            raise ApplicationError('crossbar.error.already_running', emsg)

        yield self._makers[maker_id].stop()
        del self._makers[maker_id]
        self.log.info(
            'XBR Market Maker "{maker_id}" stopped. Now running {maker_cnt} market makers in total in this worker component.',
            maker_id=maker_id,
            maker_cnt=len(self._makers))

        stopped = {
            'id': maker_id,
        }
        self.publish(u'{}.on_maker_stopped'.format(self._uri_prefix), stopped)

        returnValue(stopped)



import json

txt = """{
    "Id": "5eca2ece88c8aa2de2507466ff1dd9ec7496169002d2942a45c42c2ad3b79676",
    "Created": "2019-06-10T10:29:00.581179505Z",
    "Path": "/bin/bash",
    "Args": [],
    "State": {
        "Status": "running",
        "Running": true,
        "Paused": false,
        "Restarting": false,
        "OOMKilled": false,
        "Dead": false,
        "Pid": 3793,
        "ExitCode": 0,
        "Error": "",
        "StartedAt": "2019-06-10T10:29:02.159524647Z",
        "FinishedAt": "0001-01-01T00:00:00Z"
    },
    "Image": "sha256:d131e0fa2585a7efbfb187f70d648aa50e251d9d3b7031edf4730ca6154e221e",
    "ResolvConfPath": "/var/lib/docker/containers/5eca2ece88c8aa2de2507466ff1dd9ec7496169002d2942a45c42c2ad3b79676/resolv.conf",
    "HostnamePath": "/var/lib/docker/containers/5eca2ece88c8aa2de2507466ff1dd9ec7496169002d2942a45c42c2ad3b79676/hostname",
    "HostsPath": "/var/lib/docker/containers/5eca2ece88c8aa2de2507466ff1dd9ec7496169002d2942a45c42c2ad3b79676/hosts",
    "LogPath": "/var/lib/docker/containers/5eca2ece88c8aa2de2507466ff1dd9ec7496169002d2942a45c42c2ad3b79676/5eca2ece88c8aa2de2507466ff1dd9ec7496169002d2942a45c42c2ad3b79676-json.log",
    "Name": "/mytest",
    "RestartCount": 0,
    "Driver": "overlay2",
    "Platform": "linux",
    "MountLabel": "",
    "ProcessLabel": "",
    "AppArmorProfile": "docker-default",
    "ExecIDs": null,
    "HostConfig": {
        "Binds": [
            "/home/gareth/target:/app"
        ],
        "ContainerIDFile": "",
        "LogConfig": {
            "Type": "json-file",
            "Config": {
                "max-file": "3",
                "max-size": "10m"
            }
        },
        "NetworkMode": "default",
        "PortBindings": {},
        "RestartPolicy": {
            "Name": "no",
            "MaximumRetryCount": 0
        },
        "AutoRemove": false,
        "VolumeDriver": "",
        "VolumesFrom": null,
        "CapAdd": null,
        "CapDrop": null,
        "Dns": [],
        "DnsOptions": [],
        "DnsSearch": [],
        "ExtraHosts": null,
        "GroupAdd": null,
        "IpcMode": "shareable",
        "Cgroup": "",
        "Links": null,
        "OomScoreAdj": 0,
        "PidMode": "",
        "Privileged": false,
        "PublishAllPorts": false,
        "ReadonlyRootfs": false,
        "SecurityOpt": null,
        "UTSMode": "",
        "UsernsMode": "",
        "ShmSize": 67108864,
        "Runtime": "runc",
        "ConsoleSize": [
            0,
            0
        ],
        "Isolation": "",
        "CpuShares": 0,
        "Memory": 0,
        "NanoCpus": 0,
        "CgroupParent": "",
        "BlkioWeight": 0,
        "BlkioWeightDevice": [],
        "BlkioDeviceReadBps": null,
        "BlkioDeviceWriteBps": null,
        "BlkioDeviceReadIOps": null,
        "BlkioDeviceWriteIOps": null,
        "CpuPeriod": 0,
        "CpuQuota": 0,
        "CpuRealtimePeriod": 0,
        "CpuRealtimeRuntime": 0,
        "CpusetCpus": "",
        "CpusetMems": "",
        "Devices": [],
        "DeviceCgroupRules": null,
        "DiskQuota": 0,
        "KernelMemory": 0,
        "MemoryReservation": 0,
        "MemorySwap": 0,
        "MemorySwappiness": null,
        "OomKillDisable": false,
        "PidsLimit": 0,
        "Ulimits": null,
        "CpuCount": 0,
        "CpuPercent": 0,
        "IOMaximumIOps": 0,
        "IOMaximumBandwidth": 0,
        "MaskedPaths": [
            "/proc/asound",
            "/proc/acpi",
            "/proc/kcore",
            "/proc/keys",
            "/proc/latency_stats",
            "/proc/timer_list",
            "/proc/timer_stats",
            "/proc/sched_debug",
            "/proc/scsi",
            "/sys/firmware"
        ],
        "ReadonlyPaths": [
            "/proc/bus",
            "/proc/fs",
            "/proc/irq",
            "/proc/sys",
            "/proc/sysrq-trigger"
        ]
    },
    "GraphDriver": {
        "Data": {
            "LowerDir": "/var/lib/docker/overlay2/b27734ba84d0f0348c009649f0214309068811f78b593dd591cb40f6d4bc248f-init/diff:/var/lib/docker/overlay2/c058f6c7adc6825f992d95416cae4712d934d205304752df97af59cbe5cb33ba/diff:/var/lib/docker/overlay2/c8943979c4f38bb7f70d75c5c4f3c8731efb186007cc59ae48af40f08122ac19/diff:/var/lib/docker/overlay2/63d70bc188349a336d61c10e8e434e20d01d9e3abbb769fd6c6023d6e4e5d753/diff:/var/lib/docker/overlay2/0d8b5965209095e6a93154ba0b24077e26b2e580ce17a6a2e496dd5454554e8b/diff",
            "MergedDir": "/var/lib/docker/overlay2/b27734ba84d0f0348c009649f0214309068811f78b593dd591cb40f6d4bc248f/merged",
            "UpperDir": "/var/lib/docker/overlay2/b27734ba84d0f0348c009649f0214309068811f78b593dd591cb40f6d4bc248f/diff",
            "WorkDir": "/var/lib/docker/overlay2/b27734ba84d0f0348c009649f0214309068811f78b593dd591cb40f6d4bc248f/work"
        },
        "Name": "overlay2"
    },
    "Mounts": [
        {
            "Type": "bind",
            "Source": "/var/lib/docker/root",
            "Destination": "/app",
            "Mode": "",
            "RW": true,
            "Propagation": "rprivate"
        },
        {
            "Type": "bind",
            "Source": "/var/lib/docker/home",
            "Destination": "/home",
            "Mode": "",
            "RW": true,
            "Propagation": "rprivate"
        }
    ],
    "Config": {
        "Hostname": "5eca2ece88c8",
        "Domainname": "",
        "User": "",
        "AttachStdin": false,
        "AttachStdout": false,
        "AttachStderr": false,
        "Tty": true,
        "OpenStdin": true,
        "StdinOnce": false,
        "Env": [
            "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin"
        ],
        "Cmd": [
            "/bin/bash"
        ],
        "ArgsEscaped": true,
        "Image": "ubuntu:latest",
        "Volumes": null,
        "WorkingDir": "",
        "Entrypoint": null,
        "OnBuild": null,
        "Labels": {}
    },
    "NetworkSettings": {
        "Bridge": "",
        "SandboxID": "ae1271737d5b06c641f839b2184bdc312b46c0eed9180c3c244a832331f261fb",
        "HairpinMode": false,
        "LinkLocalIPv6Address": "",
        "LinkLocalIPv6PrefixLen": 0,
        "Ports": {},
        "SandboxKey": "/var/run/docker/netns/ae1271737d5b",
        "SecondaryIPAddresses": null,
        "SecondaryIPv6Addresses": null,
        "EndpointID": "9268246fe82700a8608484f8602989412000ce0f07ad0998ac9611670c0cff5b",
        "Gateway": "172.17.0.1",
        "GlobalIPv6Address": "",
        "GlobalIPv6PrefixLen": 0,
        "IPAddress": "172.17.0.4",
        "IPPrefixLen": 16,
        "IPv6Gateway": "",
        "MacAddress": "02:42:ac:11:00:04",
        "Networks": {
            "bridge": {
                "IPAMConfig": null,
                "Links": null,
                "Aliases": null,
                "NetworkID": "300d1064422714e93431c91e6abf32ac5e2c51a1f6ca5e69e3017187c2a7aa5d",
                "EndpointID": "9268246fe82700a8608484f8602989412000ce0f07ad0998ac9611670c0cff5b",
                "Gateway": "172.17.0.1",
                "IPAddress": "172.17.0.4",
                "IPPrefixLen": 16,
                "IPv6Gateway": "",
                "GlobalIPv6Address": "",
                "GlobalIPv6PrefixLen": 0,
                "MacAddress": "02:42:ac:11:00:04",
                "DriverOpts": null
            }
        }
    }
}"""
my_json = json.loads(txt)



"""
Helpers for test_cb_shutdown.py
"""

from twisted.internet import reactor, task


def good_join(session, details):
    print("good join: {}".format(details))

    def foo():
        return "foo"

    session.register(foo, u"foo")
    return task.deferLater(reactor, 5, lambda: None)


def failed_join(session, details):
    print("failed join: {}".format(details))

    # we wait more than 2 seconds here to avoid the "component exited
    # really quickly, shut down the node" code in
    # crossbar/node/node.py

    def fail():
        raise RuntimeError("the bad stuff")

    return task.deferLater(reactor, 5, fail)


def join_then_close(session, details):
    print("join then fail: {}".format(details))

    def fail():
        session.disconnect()

    return task.deferLater(reactor, 5, fail)



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from twisted.trial import unittest

from pytrie import StringTrie


class TestPyTrie(unittest.TestCase):
    def test_empty_tree(self):
        """
        Test trie ctor, and that is doesn't match on "any" prefix.
        """
        t = StringTrie()
        for key in ['', 'f', 'foo', 'foobar']:
            with self.assertRaises(KeyError):
                t.longest_prefix_value(key)

    def test_contains(self):
        """
        Test the contains operator.
        """
        t = StringTrie()
        test_keys = ['', 'f', 'foo', 'foobar', 'baz']
        for key in test_keys:
            t[key] = key
        for key in test_keys:
            self.assertTrue(key in t)
        for key in ['x', 'fb', 'foob', 'fooba', 'bazz']:
            self.assertFalse(key in t)

    def test_longest_prefix_1(self):
        """
        Test that keys are detected as prefix of themselfes.
        """
        t = StringTrie()
        test_keys = ['f', 'foo', 'foobar', 'baz']
        for key in test_keys:
            t[key] = key
        for key in test_keys:
            self.assertEqual(t.longest_prefix_value(key), key)

    def test_longest_prefix_2(self):
        """
        Test matching prefix lookups.
        """
        t = StringTrie()
        test_keys = ['f', 'foo', 'foobar']
        for key in test_keys:
            t[key] = key

        test_keys = {
            'foobarbaz': 'foobar',
            'foobaz': 'foo',
            'fool': 'foo',
            'foo': 'foo',
            'fob': 'f',
            'fo': 'f',
            'fx': 'f',
            'f': 'f',
        }
        for key in test_keys:
            self.assertEqual(t.longest_prefix_value(key), test_keys[key])

    def test_longest_prefix_3(self):
        """
        Test non-matching prefix lookups.
        """
        t = StringTrie()

        for key in ['x', 'fop', 'foobar']:
            t[key] = key

        for key in ['y', 'yfoo', 'fox', 'fooba']:
            with self.assertRaises(KeyError):
                t.longest_prefix_value(key)

    def test_longest_prefix_4(self):
        """
        Test that a trie with an empty string as a key contained
        matches a non-empty prefix matching lookup.
        """
        self.skip = True
        # stored_key = 'x'  # this works (and of course it should!)
        stored_key = ''  # this blows up! (and it _should_ work)
        test_key = 'xyz'

        t = StringTrie()
        t[stored_key] = stored_key
        self.assertTrue(stored_key in t)
        self.assertTrue(test_key.startswith(stored_key))
        self.assertEqual(t.longest_prefix_value(test_key), stored_key)

    # pytrie behavior is broken wrt to string keys of zero length!
    # See: https://bitbucket.org/gsakkis/pytrie/issues/4/string-keys-of-zero-length-are-not
    # We have a workaround in place for this at the relevant places.
    test_longest_prefix_4.skip = True



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

import sys
import socket
import platform

from twisted.internet import fdesc, tcp, ssl
from twisted.python.runtime import platformType

# Flag indiciating support for creating shared sockets with in-kernel
# load-balancing (!). Note that while FreeBSD had SO_REUSEPORT for ages,
# it does NOT (currently) implement load-balancing. Linux >= 3.9 and
# DragonFly BSD does.
_HAS_SHARED_LOADBALANCED_SOCKET = False

if sys.platform.startswith('linux'):
    try:
        # get Linux kernel version, like: (3, 19)
        _LINUX_KERNEL_VERSION = [int(x) for x in tuple(platform.uname()[2].split('.')[:2])]

        # SO_REUSEPORT only supported for Linux kernels >= 3.9
        if (_LINUX_KERNEL_VERSION[0] == 3 and _LINUX_KERNEL_VERSION[1] >= 9) or _LINUX_KERNEL_VERSION[0] >= 4:
            _HAS_SHARED_LOADBALANCED_SOCKET = True

            # monkey patch missing constant if needed
            if not hasattr(socket, 'SO_REUSEPORT'):
                socket.SO_REUSEPORT = 15
    except:
        pass

elif sys.platform == 'win32':
    # http://stackoverflow.com/questions/14388706/socket-options-so-reuseaddr-and-so-reuseport-how-do-they-differ-do-they-mean-t/14388707#14388707
    _HAS_SHARED_LOADBALANCED_SOCKET = True

# FIXME: DragonFly BSD claims support: http://lists.dragonflybsd.org/pipermail/commits/2013-May/130083.html

__all__ = ('create_stream_socket', 'CustomTCPPort', 'CustomTCPTLSPort')


def create_stream_socket(addressFamily, shared=False):
    """
    Create a new socket for use with Twisted's IReactor.adoptStreamPort.

    :param addressFamily: The socket address family.
    :type addressFamily: One of socket.AF_INET, socket.AF_INET6, socket.AF_UNIX
    :param shared: If `True`, request to create a shared, load-balanced socket.
                   When this feature is not available, throw an exception.
    :type shared: bool
    :returns obj -- A socket.
    """
    s = socket.socket(addressFamily, socket.SOCK_STREAM)
    s.setblocking(0)
    fdesc._setCloseOnExec(s.fileno())

    if platformType == "posix" and sys.platform != "cygwin":
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

    if shared:
        if addressFamily not in [socket.AF_INET, socket.AF_INET6]:
            raise Exception("shared sockets are only supported for IPv4 and IPv6")

        if _HAS_SHARED_LOADBALANCED_SOCKET:
            if sys.platform.startswith('linux'):
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            elif sys.platform == 'win32':
                # http://stackoverflow.com/questions/14388706/socket-options-so-reuseaddr-and-so-reuseport-how-do-they-differ-do-they-mean-t/14388707#14388707
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                # s.setsockopt(socket.SOL_SOCKET, socket.SO_EXCLUSIVEADDRUSE, 1)
            else:
                raise Exception("logic error")
        else:
            raise Exception("shared sockets unsupported on this system")

    return s


class CustomTCPPort(tcp.Port):
    """
    A custom TCP port which allows to set socket options for sharing TCP ports between multiple processes.
    """
    def __init__(self, port, factory, backlog=50, interface='', reactor=None, shared=False, user_timeout=None):
        if shared and not _HAS_SHARED_LOADBALANCED_SOCKET:
            raise Exception("shared sockets unsupported on this system")
        else:
            self._shared = shared
            self._user_timeout = user_timeout

        tcp.Port.__init__(self, port, factory, backlog, interface, reactor)

    def createInternetSocket(self):
        s = tcp.Port.createInternetSocket(self)
        if self._shared:
            if _HAS_SHARED_LOADBALANCED_SOCKET:
                if sys.platform.startswith('linux'):
                    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
                elif sys.platform == 'win32':
                    # http://stackoverflow.com/questions/14388706/socket-options-so-reuseaddr-and-so-reuseport-how-do-they-differ-do-they-mean-t/14388707#14388707
                    s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                    # s.setsockopt(socket.SOL_SOCKET, socket.SO_EXCLUSIVEADDRUSE, 1)
                else:
                    raise Exception("logic error")
            else:
                raise Exception("shared sockets unsupported on this system")

        if self._user_timeout is not None:
            s.setsockopt(socket.IPPROTO_TCP, socket.TCP_USER_TIMEOUT, self._user_timeout)

        return s


class CustomTCPTLSPort(CustomTCPPort, ssl.Port):
    """
    A custom TLS port which allows to set socket options for sharing (the underlying) TCP ports between multiple processes.
    """
    def __init__(self,
                 port,
                 factory,
                 ctxFactory,
                 backlog=50,
                 interface='',
                 reactor=None,
                 shared=False,
                 user_timeout=None):
        if shared and not _HAS_SHARED_LOADBALANCED_SOCKET:
            raise Exception("shared sockets unsupported on this system")
        else:
            self._shared = shared
            self._user_timeout = user_timeout

        ssl.Port.__init__(self, port, factory, ctxFactory, backlog, interface, reactor)



#####################################################################################
#
#  Copyright (c) Crossbar.io Technologies GmbH
#  SPDX-License-Identifier: EUPL-1.2
#
#####################################################################################

from mock import MagicMock

from twisted.trial import unittest
from twisted.internet.defer import Deferred
from twisted.internet import task

# WebSocket protocol gets used below, and the latter
# calls txaio.make_logger(). If we don't explicitly select
# the network framework before, we get an exception
# "To use txaio, you must first select a framework" from txaio
import txaio

txaio.use_twisted()  # noqa

from crossbar.node.node import NodeController


class CleanupHandler(unittest.TestCase):
    def setUp(self):
        self.transport = MagicMock()
        self.worker = MagicMock()
        self.worker.proto.transport = self.transport
        self.worker.pid = '42'
        self.worker.ready = Deferred()
        self.worker.exit = Deferred()

    def test_kill_after_term(self):
        reactor = task.Clock()
        NodeController._cleanup_worker(reactor, self.worker)

        # should have sent TERM now
        calls = self.worker.proto.transport.method_calls
        self.assertTrue(calls[0][0] == "signalProcess")
        self.assertTrue(calls[0][1] == ('TERM', ))

        # skip ahead until our KILL. we loop because we only run one
        # timed-out thing on each advance maybe? Anyway it runs
        # timeout() only twice if I advance(30) here instead...
        for x in range(30):
            reactor.advance(1)

        calls = self.worker.proto.transport.method_calls
        self.assertTrue(calls[1][0] == "signalProcess")
        self.assertTrue(calls[1][1] == ("KILL", ))



##############################################################################
#
#                        Crossbar.io
#     Copyright (C) Crossbar.io Technologies GmbH. All rights reserved.
#
##############################################################################

import datetime
from collections.abc import Mapping, Sequence

import six
import txaio
from twisted.internet.defer import inlineCallbacks, returnValue

from autobahn.util import utcstr
from autobahn.wamp.exception import ApplicationError
from crossbar.common import checkconfig
from crossbar.common.checkconfig import get_config_value

try:
    from txpostgres import txpostgres
    _HAS_POSTGRESQL = True
except ImportError:
    _HAS_POSTGRESQL = False

__all__ = ('PostgresConnectionPool', )

if _HAS_POSTGRESQL:

    class PostgreSQLConnection(object):
        """
        A PostgreSQL database connection pool.
        """

        log = txaio.make_logger()

        def __init__(self, id, config):
            """

            :param id: The ID of the connection item.
            :type id: unicode
            :param config: The connection item configuration.
            :type config: dict
            """
            self.id = id
            self.config = config
            self.started = None
            self.stopped = None

            params = {
                'host': config.get('host', 'localhost'),
                'port': config.get('port', 5432),
                'database': config['database'],
                'user': config['user'],
                'password': get_config_value(config, 'password'),
            }
            self.pool = txpostgres.ConnectionPool(None, min=5, **params)

        def start(self):
            self.started = datetime.utcnow()
            return self.pool.start()

        def stop(self):
            self.stopped = datetime.utcnow()
            return self.pool.close()

        def marshal(self):
            return {
                u'id': self.id,
                u'started': utcstr(self.started),
                u'stopped': utcstr(self.stopped) if self.stopped else None,
                u'config': self.config,
            }


class PostgresConnectionPool(object):

    log = txaio.make_logger()

    def __init__(self, personality, session):
        self._personality = personality
        self._session = session
        self._connections = {}

    @staticmethod
    def check_connection(personality, connection, ignore=[]):
        """
        Check a connection item (such as a PostgreSQL or Oracle database connection pool).
        """
        if 'id' in connection:
            checkconfig.check_id(connection['id'])

        if 'type' not in connection:
            raise checkconfig.InvalidConfigException("missing mandatory attribute 'type' in connection configuration")

        valid_types = ['postgres']
        if connection['type'] not in valid_types:
            raise checkconfig.InvalidConfigException(
                "invalid type '{}' for connection type - must be one of {}".format(connection['type'], valid_types))

        if connection['type'] == 'postgres':
            checkconfig.check_dict_args(
                {
                    'id': (False, [six.text_type]),
                    'type': (True, [six.text_type]),
                    'host': (False, [six.text_type]),
                    'port': (False, six.integer_types),
                    'database': (True, [six.text_type]),
                    'user': (True, [six.text_type]),
                    'password': (False, [six.text_type]),
                    'options': (False, [Mapping]),
                }, connection, "PostgreSQL connection configuration")

            if 'port' in connection:
                checkconfig.check_endpoint_port(connection['port'])

            if 'options' in connection:
                checkconfig.check_dict_args(
                    {
                        'min_connections': (False, six.integer_types),
                        'max_connections': (False, six.integer_types),
                    }, connection['options'], "PostgreSQL connection options")

        else:
            raise checkconfig.InvalidConfigException('logic error')

    @staticmethod
    def check_connections(personality, connections):
        """
        Connections can be present in controller, router and container processes.
        """
        if not isinstance(connections, Sequence):
            raise checkconfig.InvalidConfigException("'connections' items must be lists ({} encountered)".format(
                type(connections)))

        for i, connection in enumerate(connections):
            personality.check_connection(personality, connection)

    @inlineCallbacks
    def start_connection(self, id, config, details=None):
        """
        Starts a connection in this process.

        :param id: The ID for the started connection.
        :type id: unicode

        :param config: Connection configuration.
        :type config: dict

        :param details: Caller details.
        :type details: instance of :class:`autobahn.wamp.types.CallDetails`

        :returns: The connection.
        """
        self.log.debug("start_connection: id={id}, config={config}", id=id, config=config)

        # prohibit starting a component twice
        #
        if id in self._connections:
            emsg = "cannot start connection: a connection with id={} is already started".format(id)
            self.log.warn(emsg)
            raise ApplicationError(u"crossbar.error.invalid_configuration", emsg)

        # check configuration
        #
        try:
            self._personality.check_connection(self._personality, config)
        except Exception as e:
            emsg = "invalid connection configuration ({})".format(e)
            self.log.warn(emsg)
            raise ApplicationError(u"crossbar.error.invalid_configuration", emsg)
        else:
            self.log.info("Starting {ptype} in process.", ptype=config['type'])

        if config['type'] == u'postgresql.connection':
            if _HAS_POSTGRESQL:
                connection = PostgreSQLConnection(id, config)
            else:
                emsg = "unable to start connection - required PostgreSQL driver package not installed"
                self.log.warn(emsg)
                raise ApplicationError(u"crossbar.error.feature_unavailable", emsg)
        else:
            # should not arrive here
            raise Exception("logic error")

        self._connections[id] = connection

        try:
            yield connection.start()
            self.log.info("Connection {connection_type} started '{connection_id}'",
                          connection_id=id,
                          connection_type=config['type'])
        except:
            del self._connections[id]
            raise

        state = connection.marshal()

        self._session.publish(u'crossbar.node.process.on_connection_start', state)

        returnValue(state)

    @inlineCallbacks
    def stop_connection(self, id, details=None):
        """
        Stop a connection currently running within this process.

        :param id: The ID of the connection to stop.
        :type id: unicode

        :param details: Caller details.
        :type details: instance of :class:`autobahn.wamp.types.CallDetails`

        :returns: A dict with component start information.
        """
        self.log.debug("stop_connection: id={id}", id=id)

        if id not in self._connections:
            raise ApplicationError(u'crossbar.error.no_such_object',
                                   'no connection with ID {} running in this process'.format(id))

        connection = self._connections[id]

        try:
            yield connection.stop()
        except Exception as e:
            self.log.warn('could not stop connection {id}: {error}', error=e)
            raise

        del self._connections[id]

        state = connection.marshal()

        self._session.publish(u'crossbar.node.process.on_connection_stop', state)

        returnValue(state)

    def get_connections(self, details=None):
        """
        Get connections currently running within this processs.

        :param details: Caller details.
        :type details: instance of :class:`autobahn.wamp.types.CallDetails`

        :returns: List of connections.
        """
        self.log.debug("get_connections")

        res = []
        for c in self._connections.values():
            res.append(c.marshal())
        return res



###############################################################################
#
# Crossbar.io Master
# Copyright (c) Crossbar.io Technologies GmbH. Licensed under EUPLv1.2.
#
###############################################################################

from typing import Dict, List  # noqa

from crossbar.master.api.remote import RemoteApi

__all__ = ('RemoteNativeProcessApi', )


class RemoteNativeProcessApi(RemoteApi):
    """
    Remote API to CF native node processes:

    * node controller
    * router worker
    * container worker
    * proxy worker
    """

    PREFIX = u'crossbarfabriccenter.remote.nativeprocess.'

    PROCS = {
        # these are node level procedures
        u'node': [
            u'get_cpu_count',
            u'get_cpu_affinity',
            u'set_cpu_affinity',
            u'get_process_info',
            u'get_process_stats',
            u'set_process_stats_monitoring',
            u'get_process_monitor',
            u'trigger_gc',
            u'start_manhole',
            u'stop_manhole',
            u'get_manhole',
        ],
    }

    EVENTS = {
        # these are node level topics
        u'node': []
    }  # type: Dict[str, List]
